# SPDX-License-Identifier: BSD-2-Clause

import warnings
from enum          import Enum

from torii.hdl.ast import *

from ..utils       import ToriiTestSuiteCase


class UnsignedEnum(Enum):
	FOO = 1
	BAR = 2
	BAZ = 3


class SignedEnum(Enum):
	FOO = -1
	BAR = 0
	BAZ = +1


class StringEnum(Enum):
	FOO = 'a'
	BAR = 'b'


class TypedEnum(int, Enum):
	FOO = 1
	BAR = 2
	BAZ = 3


class ShapeTestCase(ToriiTestSuiteCase):
	def test_make(self):
		s1 = Shape()
		self.assertEqual(s1.width, 1)
		self.assertEqual(s1.signed, False)
		s2 = Shape(signed = True)
		self.assertEqual(s2.width, 1)
		self.assertEqual(s2.signed, True)
		s3 = Shape(3, True)
		self.assertEqual(s3.width, 3)
		self.assertEqual(s3.signed, True)
		s4 = Shape(0)
		self.assertEqual(s4.width, 0)
		self.assertEqual(s4.signed, False)

	def test_make_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Width must be an integer, not \'a\'$'
		):
			Shape('a')
		with self.assertRaisesRegex(
			ValueError,
			r'^Width of an unsigned value must be zero or a positive integer, not -1$'
		):
			Shape(-1, signed = False)
		with self.assertRaisesRegex(
			ValueError,
			r'^Width of a signed value must be a positive integer, not 0$'
		):
			Shape(0, signed = True)

	def test_compare_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Shapes may be compared with shape-castable objects, not \'hi\'$'
		):
			Shape(1, True) == 'hi'

	def test_compare_tuple_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Shapes may be compared with shape-castable objects, not \(2, 3\)$'
		):
			Shape(1, True) == (2, 3)

	def test_repr(self):
		self.assertEqual(repr(Shape()), 'unsigned(1)')
		self.assertEqual(repr(Shape(2, True)), 'signed(2)')

	def test_convert_tuple_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^cannot unpack non-iterable Shape object$'
		):
			width, signed = Shape()

	def test_unsigned(self):
		s1 = unsigned(2)
		self.assertIsInstance(s1, Shape)
		self.assertEqual(s1.width, 2)
		self.assertEqual(s1.signed, False)

	def test_signed(self):
		s1 = signed(2)
		self.assertIsInstance(s1, Shape)
		self.assertEqual(s1.width, 2)
		self.assertEqual(s1.signed, True)

	def test_cast_shape(self):
		s1 = Shape.cast(unsigned(1))
		self.assertEqual(s1.width, 1)
		self.assertEqual(s1.signed, False)
		s2 = Shape.cast(signed(3))
		self.assertEqual(s2.width, 3)
		self.assertEqual(s2.signed, True)

	def test_cast_int(self):
		s1 = Shape.cast(2)
		self.assertEqual(s1.width, 2)
		self.assertEqual(s1.signed, False)

	def test_cast_int_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Width of an unsigned value must be zero or a positive integer, not -1$'
		):
			Shape.cast(-1)

	def test_cast_tuple_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Object \(-1, True\) cannot be converted to a Torii shape$'
		):
			Shape.cast((-1, True))

	def test_cast_range(self):
		s1 = Shape.cast(range(0, 8))
		self.assertEqual(s1.width, 3)
		self.assertEqual(s1.signed, False)
		s2 = Shape.cast(range(0, 9))
		self.assertEqual(s2.width, 4)
		self.assertEqual(s2.signed, False)
		s3 = Shape.cast(range(-7, 8))
		self.assertEqual(s3.width, 4)
		self.assertEqual(s3.signed, True)
		s4 = Shape.cast(range(0, 1))
		self.assertEqual(s4.width, 1)
		self.assertEqual(s4.signed, False)
		s5 = Shape.cast(range(-1, 0))
		self.assertEqual(s5.width, 1)
		self.assertEqual(s5.signed, True)
		s6 = Shape.cast(range(0, 0))
		self.assertEqual(s6.width, 0)
		self.assertEqual(s6.signed, False)
		s7 = Shape.cast(range(-1, -1))
		self.assertEqual(s7.width, 0)
		self.assertEqual(s7.signed, False)
		s8 = Shape.cast(range(0, 10, 3))
		self.assertEqual(s8.width, 4)
		self.assertEqual(s8.signed, False)

	def test_cast_enum(self):
		s1 = Shape.cast(UnsignedEnum)
		self.assertEqual(s1.width, 2)
		self.assertEqual(s1.signed, False)
		s2 = Shape.cast(SignedEnum)
		self.assertEqual(s2.width, 2)
		self.assertEqual(s2.signed, True)

	def test_cast_enum_bad(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Only enumerations whose members have constant-castable values can be used in Torii code$'
		):
			Shape.cast(StringEnum)

	def test_cast_bad(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Object \'foo\' cannot be converted to a Torii shape$'
		):
			Shape.cast('foo')


class MockShapeCastable(ShapeCastable):
	def __init__(self, dest):
		self.dest = dest

	def as_shape(self):
		return self.dest


class ShapeCastableTestCase(ToriiTestSuiteCase):
	def test_no_override(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Class \'MockShapeCastableNoOverride\' deriving from `ShapeCastable` must '
			r'override the `as_shape` method$'
		):
			class MockShapeCastableNoOverride(ShapeCastable):
				def __init__(self):
					pass

	def test_cast(self):
		sc = MockShapeCastable(unsigned(2))
		self.assertEqual(Shape.cast(sc), unsigned(2))

	def test_recurse_bad(self):
		sc = MockShapeCastable(None)
		sc.dest = sc
		with self.assertRaisesRegex(
			RecursionError,
			r'^Shape-castable object <.+> casts to itself$'
		):
			Shape.cast(sc)

	def test_recurse(self):
		sc = MockShapeCastable(MockShapeCastable(unsigned(1)))
		self.assertEqual(Shape.cast(sc), unsigned(1))


class ValueTestCase(ToriiTestSuiteCase):
	def test_cast(self):
		self.assertIsInstance(Value.cast(0), Const)
		self.assertIsInstance(Value.cast(True), Const)
		c = Const(0)
		self.assertIs(Value.cast(c), c)
		with self.assertRaisesRegex(
			TypeError,
			r'^Object \'str\' cannot be converted to a Torii value$'
		):
			Value.cast('str')

	def test_cast_enum(self):
		e1 = Value.cast(UnsignedEnum.FOO)
		self.assertIsInstance(e1, Const)
		self.assertEqual(e1.shape(), unsigned(2))
		e2 = Value.cast(SignedEnum.FOO)
		self.assertIsInstance(e2, Const)
		self.assertEqual(e2.shape(), signed(2))

	def test_cast_typedenum(self):
		e1 = Value.cast(TypedEnum.FOO)
		self.assertIsInstance(e1, Const)
		self.assertEqual(e1.shape(), unsigned(2))

	def test_cast_enum_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Only enumerations whose members have constant-castable values can be used in Torii code$'
		):
			Value.cast(StringEnum.FOO)

	def test_bool(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Attempted to convert Torii value to Python boolean$'
		):
			if Const(0):
				pass

	def test_len(self):
		self.assertEqual(len(Const(10)), 4)

	def test_getitem_int(self):
		s1 = Const(10)[0]
		self.assertIsInstance(s1, Slice)
		self.assertEqual(s1.start, 0)
		self.assertEqual(s1.stop, 1)
		s2 = Const(10)[-1]
		self.assertIsInstance(s2, Slice)
		self.assertEqual(s2.start, 3)
		self.assertEqual(s2.stop, 4)
		with self.assertRaisesRegex(
			IndexError,
			r'^Index 5 is out of bounds for a 4-bit value$'
		):
			Const(10)[5]

	def test_getitem_slice(self):
		s1 = Const(10)[1:3]
		self.assertIsInstance(s1, Slice)
		self.assertEqual(s1.start, 1)
		self.assertEqual(s1.stop, 3)
		s2 = Const(10)[1:-2]
		self.assertIsInstance(s2, Slice)
		self.assertEqual(s2.start, 1)
		self.assertEqual(s2.stop, 2)
		s3 = Const(31)[::2]
		self.assertIsInstance(s3, Cat)
		self.assertIsInstance(s3.parts[0], Slice)
		self.assertEqual(s3.parts[0].start, 0)
		self.assertEqual(s3.parts[0].stop, 1)
		self.assertIsInstance(s3.parts[1], Slice)
		self.assertEqual(s3.parts[1].start, 2)
		self.assertEqual(s3.parts[1].stop, 3)
		self.assertIsInstance(s3.parts[2], Slice)
		self.assertEqual(s3.parts[2].start, 4)
		self.assertEqual(s3.parts[2].stop, 5)

	def test_getitem_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Cannot index value with \'str\'$'
		):
			Const(31)['str']

	def test_shift_left(self):
		self.assertRepr(
			Const(256, unsigned(9)).shift_left(0),
			'(cat (const 0\'d0) (const 9\'d256))'
		)

		self.assertRepr(
			Const(256, unsigned(9)).shift_left(1),
			'(cat (const 1\'d0) (const 9\'d256))'
		)
		self.assertRepr(
			Const(256, unsigned(9)).shift_left(5),
			'(cat (const 5\'d0) (const 9\'d256))'
		)
		self.assertRepr(
			Const(256, signed(9)).shift_left(1),
			'(s (cat (const 1\'d0) (const 9\'sd-256)))'
		)
		self.assertRepr(
			Const(256, signed(9)).shift_left(5),
			'(s (cat (const 5\'d0) (const 9\'sd-256)))'
		)

		self.assertRepr(
			Const(256, unsigned(9)).shift_left(-1),
			'(slice (const 9\'d256) 1:9)'
		)
		self.assertRepr(
			Const(256, unsigned(9)).shift_left(-5),
			'(slice (const 9\'d256) 5:9)'
		)
		self.assertRepr(
			Const(256, signed(9)).shift_left(-1),
			'(s (slice (const 9\'sd-256) 1:9))'
		)
		self.assertRepr(
			Const(256, signed(9)).shift_left(-5),
			'(s (slice (const 9\'sd-256) 5:9))'
		)
		self.assertRepr(
			Const(256, signed(9)).shift_left(-15),
			'(s (slice (const 9\'sd-256) 9:9))'
		)

	def test_shift_left_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Shift amount must be an integer, not \'str\'$'
		):
			Const(31).shift_left('str')

	def test_shift_right(self):
		self.assertRepr(
			Const(256, unsigned(9)).shift_right(0),
			'(slice (const 9\'d256) 0:9)'
		)

		self.assertRepr(
			Const(256, unsigned(9)).shift_right(-1),
			'(cat (const 1\'d0) (const 9\'d256))'
		)
		self.assertRepr(
			Const(256, unsigned(9)).shift_right(-5),
			'(cat (const 5\'d0) (const 9\'d256))'
		)
		self.assertRepr(
			Const(256, signed(9)).shift_right(-1),
			'(s (cat (const 1\'d0) (const 9\'sd-256)))'
		)
		self.assertRepr(
			Const(256, signed(9)).shift_right(-5),
			'(s (cat (const 5\'d0) (const 9\'sd-256)))'
		)

		self.assertRepr(
			Const(256, unsigned(9)).shift_right(1),
			'(slice (const 9\'d256) 1:9)'
		)
		self.assertRepr(
			Const(256, unsigned(9)).shift_right(5),
			'(slice (const 9\'d256) 5:9)'
		)
		self.assertRepr(
			Const(256, signed(9)).shift_right(1),
			'(s (slice (const 9\'sd-256) 1:9))'
		)
		self.assertRepr(
			Const(256, signed(9)).shift_right(5),
			'(s (slice (const 9\'sd-256) 5:9))'
		)
		self.assertRepr(
			Const(256, signed(9)).shift_right(15),
			'(s (slice (const 9\'sd-256) 9:9))'
		)

	def test_shift_right_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Shift amount must be an integer, not \'str\'$'
		):
			Const(31).shift_left('str')

	def test_rotate_left(self):
		self.assertRepr(
			Const(256).rotate_left(1),
			'(cat (slice (const 9\'d256) 8:9) (slice (const 9\'d256) 0:8))'
		)
		self.assertRepr(
			Const(256).rotate_left(7),
			'(cat (slice (const 9\'d256) 2:9) (slice (const 9\'d256) 0:2))'
		)
		self.assertRepr(
			Const(256).rotate_left(-1),
			'(cat (slice (const 9\'d256) 1:9) (slice (const 9\'d256) 0:1))'
		)
		self.assertRepr(
			Const(256).rotate_left(-7),
			'(cat (slice (const 9\'d256) 7:9) (slice (const 9\'d256) 0:7))'
		)
		self.assertRepr(
			Const(0, 0).rotate_left(3),
			'(cat (slice (const 0\'d0) 0:0) (slice (const 0\'d0) 0:0))'
		)
		self.assertRepr(
			Const(0, 0).rotate_left(-3),
			'(cat (slice (const 0\'d0) 0:0) (slice (const 0\'d0) 0:0))'
		)


	def test_rotate_left_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Rotate amount must be an integer, not \'str\'$'
		):
			Const(31).rotate_left('str')

	def test_rotate_right(self):
		self.assertRepr(
			Const(256).rotate_right(1),
			'(cat (slice (const 9\'d256) 1:9) (slice (const 9\'d256) 0:1))'
		)
		self.assertRepr(
			Const(256).rotate_right(7),
			'(cat (slice (const 9\'d256) 7:9) (slice (const 9\'d256) 0:7))'
		)
		self.assertRepr(
			Const(256).rotate_right(-1),
			'(cat (slice (const 9\'d256) 8:9) (slice (const 9\'d256) 0:8))'
		)
		self.assertRepr(
			Const(256).rotate_right(-7),
			'(cat (slice (const 9\'d256) 2:9) (slice (const 9\'d256) 0:2))'
		)
		self.assertRepr(
			Const(0, 0).rotate_right(3),
			'(cat (slice (const 0\'d0) 0:0) (slice (const 0\'d0) 0:0))'
		)
		self.assertRepr(
			Const(0, 0).rotate_right(-3),
			'(cat (slice (const 0\'d0) 0:0) (slice (const 0\'d0) 0:0))'
		)


	def test_rotate_right_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Rotate amount must be an integer, not \'str\'$'
		):
			Const(31).rotate_right('str')

	def test_replicate_shape(self):
		s1 = Const(10).replicate(3)
		self.assertEqual(s1.shape(), unsigned(12))
		self.assertIsInstance(s1.shape(), Shape)
		s2 = Const(10).replicate(0)
		self.assertEqual(s2.shape(), unsigned(0))

	def test_replicate_count_wrong(self):
		with self.assertRaises(TypeError):
			Const(10).replicate(-1)
		with self.assertRaises(TypeError):
			Const(10).replicate('str')

	def test_replicate_repr(self):
		s = Const(10).replicate(3)
		self.assertEqual(repr(s), '(cat (const 4\'d10) (const 4\'d10) (const 4\'d10))')


class ConstTestCase(ToriiTestSuiteCase):
	def test_shape(self):
		self.assertEqual(Const(0).shape(),   unsigned(1))
		self.assertIsInstance(Const(0).shape(), Shape)
		self.assertEqual(Const(1).shape(),   unsigned(1))
		self.assertEqual(Const(10).shape(),  unsigned(4))
		self.assertEqual(Const(-10).shape(), signed(5))

		self.assertEqual(Const(1, 4).shape(),          unsigned(4))
		self.assertEqual(Const(-1, 4).shape(),         signed(4))
		self.assertEqual(Const(1, signed(4)).shape(),  signed(4))
		self.assertEqual(Const(0, unsigned(0)).shape(), unsigned(0))

	def test_shape_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Width of an unsigned value must be zero or a positive integer, not -1$'
		):
			Const(1, -1)

	def test_wrong_fencepost(self):
		with self.assertWarnsRegex(
			SyntaxWarning,
			r'^Value 10 equals the non-inclusive end of the constant shape '
			r'range\(0, 10\); this is likely an off-by-one error$'
		):
			Const(10, range(10))

	def test_normalization(self):
		self.assertEqual(Const(0b10110, signed(5)).value, -10)
		self.assertEqual(Const(0b10000, signed(4)).value, 0)
		self.assertEqual(Const(-16, 4).value, 0)

	def test_value(self):
		self.assertEqual(Const(10).value, 10)

	def test_repr(self):
		self.assertEqual(repr(Const(10)),  '(const 4\'d10)')
		self.assertEqual(repr(Const(-10)), '(const 5\'sd-10)')

	def test_hash(self):
		with self.assertRaises(TypeError):
			hash(Const(0))


class OperatorTestCase(ToriiTestSuiteCase):
	def test_bool(self):
		v = Const(0, 4).bool()
		self.assertEqual(repr(v), '(b (const 4\'d0))')
		self.assertEqual(v.shape(), unsigned(1))

	def test_invert(self):
		v = ~Const(0, 4)
		self.assertEqual(repr(v), '(~ (const 4\'d0))')
		self.assertEqual(v.shape(), unsigned(4))

	def test_as_unsigned(self):
		v = Const(-1, signed(4)).as_unsigned()
		self.assertEqual(repr(v), '(u (const 4\'sd-1))')
		self.assertEqual(v.shape(), unsigned(4))

	def test_as_signed(self):
		v = Const(1, unsigned(4)).as_signed()
		self.assertEqual(repr(v), '(s (const 4\'d1))')
		self.assertEqual(v.shape(), signed(4))

	def test_pos(self):
		self.assertRepr(+Const(10), '(const 4\'d10)')

	def test_neg(self):
		v1 = -Const(0, unsigned(4))
		self.assertEqual(repr(v1), '(- (const 4\'d0))')
		self.assertEqual(v1.shape(), signed(5))
		v2 = -Const(0, signed(4))
		self.assertEqual(repr(v2), '(- (const 4\'sd0))')
		self.assertEqual(v2.shape(), signed(5))

	def test_add(self):
		v1 = Const(0, unsigned(4)) + Const(0, unsigned(6))
		self.assertEqual(repr(v1), '(+ (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v1.shape(), unsigned(7))
		v2 = Const(0, signed(4)) + Const(0, signed(6))
		self.assertEqual(v2.shape(), signed(7))
		v3 = Const(0, signed(4)) + Const(0, unsigned(4))
		self.assertEqual(v3.shape(), signed(6))
		v4 = Const(0, unsigned(4)) + Const(0, signed(4))
		self.assertEqual(v4.shape(), signed(6))
		v5 = 10 + Const(0, 4)
		self.assertEqual(v5.shape(), unsigned(5))

	def test_sub(self):
		v1 = Const(0, unsigned(4)) - Const(0, unsigned(6))
		self.assertEqual(repr(v1), '(- (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v1.shape(), signed(7))
		v2 = Const(0, signed(4)) - Const(0, signed(6))
		self.assertEqual(v2.shape(), signed(7))
		v3 = Const(0, signed(4)) - Const(0, unsigned(4))
		self.assertEqual(v3.shape(), signed(6))
		v4 = Const(0, unsigned(4)) - Const(0, signed(4))
		self.assertEqual(v4.shape(), signed(6))
		v5 = 10 - Const(0, 4)
		self.assertEqual(v5.shape(), signed(5))
		v6 = 1 - Const(2)
		self.assertEqual(v6.shape(), signed(3))

	def test_mul(self):
		v1 = Const(0, unsigned(4)) * Const(0, unsigned(6))
		self.assertEqual(repr(v1), '(* (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v1.shape(), unsigned(10))
		v2 = Const(0, signed(4)) * Const(0, signed(6))
		self.assertEqual(v2.shape(), signed(10))
		v3 = Const(0, signed(4)) * Const(0, unsigned(4))
		self.assertEqual(v3.shape(), signed(8))
		v5 = 10 * Const(0, 4)
		self.assertEqual(v5.shape(), unsigned(8))

	def test_mod(self):
		v1 = Const(0, unsigned(4)) % Const(0, unsigned(6))
		self.assertEqual(repr(v1), '(% (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v1.shape(), unsigned(6))
		v3 = Const(0, signed(4)) % Const(0, unsigned(4))
		self.assertEqual(v3.shape(), unsigned(4))
		v4 = Const(0, signed(4)) % Const(0, signed(6))
		self.assertEqual(v4.shape(), signed(6))
		v5 = 10 % Const(0, 4)
		self.assertEqual(v5.shape(), unsigned(4))

	def test_floordiv(self):
		v1 = Const(0, unsigned(4)) // Const(0, unsigned(6))
		self.assertEqual(repr(v1), '(// (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v1.shape(), unsigned(4))
		v3 = Const(0, signed(4)) // Const(0, unsigned(4))
		self.assertEqual(v3.shape(), signed(4))
		v4 = Const(0, signed(4)) // Const(0, signed(6))
		self.assertEqual(v4.shape(), signed(5))
		v5 = 10 // Const(0, 4)
		self.assertEqual(v5.shape(), unsigned(4))

	def test_and(self):
		v1 = Const(0, unsigned(4)) & Const(0, unsigned(6))
		self.assertEqual(repr(v1), '(& (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v1.shape(), unsigned(6))
		v2 = Const(0, signed(4)) & Const(0, signed(6))
		self.assertEqual(v2.shape(), signed(6))
		v3 = Const(0, signed(4)) & Const(0, unsigned(4))
		self.assertEqual(v3.shape(), signed(5))
		v4 = Const(0, unsigned(4)) & Const(0, signed(4))
		self.assertEqual(v4.shape(), signed(5))
		v5 = 10 & Const(0, 4)
		self.assertEqual(v5.shape(), unsigned(4))

	def test_or(self):
		v1 = Const(0, unsigned(4)) | Const(0, unsigned(6))
		self.assertEqual(repr(v1), '(| (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v1.shape(), unsigned(6))
		v2 = Const(0, signed(4)) | Const(0, signed(6))
		self.assertEqual(v2.shape(), signed(6))
		v3 = Const(0, signed(4)) | Const(0, unsigned(4))
		self.assertEqual(v3.shape(), signed(5))
		v4 = Const(0, unsigned(4)) | Const(0, signed(4))
		self.assertEqual(v4.shape(), signed(5))
		v5 = 10 | Const(0, 4)
		self.assertEqual(v5.shape(), unsigned(4))

	def test_xor(self):
		v1 = Const(0, unsigned(4)) ^ Const(0, unsigned(6))
		self.assertEqual(repr(v1), '(^ (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v1.shape(), unsigned(6))
		v2 = Const(0, signed(4)) ^ Const(0, signed(6))
		self.assertEqual(v2.shape(), signed(6))
		v3 = Const(0, signed(4)) ^ Const(0, unsigned(4))
		self.assertEqual(v3.shape(), signed(5))
		v4 = Const(0, unsigned(4)) ^ Const(0, signed(4))
		self.assertEqual(v4.shape(), signed(5))
		v5 = 10 ^ Const(0, 4)
		self.assertEqual(v5.shape(), unsigned(4))

	def test_shl(self):
		v1 = Const(1, 4) << Const(4)
		self.assertEqual(repr(v1), '(<< (const 4\'d1) (const 3\'d4))')
		self.assertEqual(v1.shape(), unsigned(11))

	def test_shl_wrong(self):
		with self.assertRaisesRegex(TypeError, r'^Shift amount must be unsigned$'):
			1 << Const(0, signed(6))
		with self.assertRaisesRegex(TypeError, r'^Shift amount must be unsigned$'):
			Const(1, unsigned(4)) << -1

	def test_shr(self):
		v1 = Const(1, 4) >> Const(4)
		self.assertEqual(repr(v1), '(>> (const 4\'d1) (const 3\'d4))')
		self.assertEqual(v1.shape(), unsigned(4))

	def test_shr_wrong(self):
		with self.assertRaisesRegex(TypeError, r'^Shift amount must be unsigned$'):
			1 << Const(0, signed(6))
		with self.assertRaisesRegex(TypeError, r'^Shift amount must be unsigned$'):
			Const(1, unsigned(4)) << -1

	def test_lt(self):
		v = Const(0, 4) < Const(0, 6)
		self.assertEqual(repr(v), '(< (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v.shape(), unsigned(1))

	def test_le(self):
		v = Const(0, 4) <= Const(0, 6)
		self.assertEqual(repr(v), '(<= (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v.shape(), unsigned(1))

	def test_gt(self):
		v = Const(0, 4) > Const(0, 6)
		self.assertEqual(repr(v), '(> (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v.shape(), unsigned(1))

	def test_ge(self):
		v = Const(0, 4) >= Const(0, 6)
		self.assertEqual(repr(v), '(>= (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v.shape(), unsigned(1))

	def test_eq(self):
		v = Const(0, 4) == Const(0, 6)
		self.assertEqual(repr(v), '(== (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v.shape(), unsigned(1))

	def test_ne(self):
		v = Const(0, 4) != Const(0, 6)
		self.assertEqual(repr(v), '(!= (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v.shape(), unsigned(1))

	def test_mux(self):
		s  = Const(0)
		v1 = Mux(s, Const(0, unsigned(4)), Const(0, unsigned(6)))
		self.assertEqual(repr(v1), '(m (const 1\'d0) (const 4\'d0) (const 6\'d0))')
		self.assertEqual(v1.shape(), unsigned(6))
		v2 = Mux(s, Const(0, signed(4)), Const(0, signed(6)))
		self.assertEqual(v2.shape(), signed(6))
		v3 = Mux(s, Const(0, signed(4)), Const(0, unsigned(4)))
		self.assertEqual(v3.shape(), signed(5))
		v4 = Mux(s, Const(0, unsigned(4)), Const(0, signed(4)))
		self.assertEqual(v4.shape(), signed(5))

	def test_mux_wide(self):
		s = Const(0b100)
		v = Mux(s, Const(0, unsigned(4)), Const(0, unsigned(6)))
		self.assertEqual(repr(v), '(m (const 3\'d4) (const 4\'d0) (const 6\'d0))')

	def test_mux_bool(self):
		v = Mux(True, Const(0), Const(0))
		self.assertEqual(repr(v), '(m (const 1\'d1) (const 1\'d0) (const 1\'d0))')

	def test_any(self):
		v = Const(0b101).any()
		self.assertEqual(repr(v), '(r| (const 3\'d5))')

	def test_all(self):
		v = Const(0b101).all()
		self.assertEqual(repr(v), '(r& (const 3\'d5))')

	def test_xor_value(self):
		v = Const(0b101).xor()
		self.assertEqual(repr(v), '(r^ (const 3\'d5))')

	def test_matches(self):
		s = Signal(4)
		self.assertRepr(s.matches(), '(const 1\'d1)')
		self.assertRepr(s.matches(1), '''
		(== (sig s) (const 1'd1))
		''')
		self.assertRepr(s.matches(0, 1), '''
		(r| (cat (== (sig s) (const 1'd0)) (== (sig s) (const 1'd1))))
		''')
		self.assertRepr(s.matches('10--'), '''
		(== (& (sig s) (const 4'd12)) (const 4'd8))
		''')
		self.assertRepr(s.matches('1 0--'), '''
		(== (& (sig s) (const 4'd12)) (const 4'd8))
		''')

	def test_matches_enum(self):
		s = Signal(SignedEnum)
		self.assertRepr(s.matches(SignedEnum.FOO), '''
		(== (sig s) (const 1'sd-1))
		''')

	def test_matches_width_wrong(self):
		s = Signal(4)
		with self.assertRaisesRegex(
			SyntaxError,
			r'^Match pattern \'--\' must have the same width as match value \(which is 4\)$'
		):
			s.matches('--')
		with self.assertWarnsRegex(
			SyntaxWarning, (
				r'^Match pattern \'10110\' is wider than match value \(which has width 4\); '
				r'comparison will never be true$'
			)
		):
			s.matches(0b10110)

	def test_matches_bits_wrong(self):
		s = Signal(4)
		with self.assertRaisesRegex(
			SyntaxError, (
				r'^Match pattern \'abc\' must consist of 0, 1, and - \(don\'t care\) bits, '
				r'and may include whitespace$'
			)
		):
			s.matches('abc')

	def test_matches_pattern_wrong(self):
		s = Signal(4)
		with self.assertRaisesRegex(
			SyntaxError,
			r'^Match pattern must be an integer, a string, or an enumeration, not 1\.0$'
		):
			s.matches(1.0)

	def test_hash(self):
		with self.assertRaises(TypeError):
			hash(Const(0) + Const(0))

	def test_abs(self):
		s = Signal(4)
		self.assertRepr(abs(s), '(sig s)')
		s = Signal(signed(4))
		self.assertRepr(
			abs(s),
			'(slice (m (>= (sig s) (const 1\'d0)) (sig s) (- (sig s))) 0:4)'
		)
		self.assertEqual(abs(s).shape(), unsigned(4))


class SliceTestCase(ToriiTestSuiteCase):
	def test_shape(self):
		s1 = Const(10)[2]
		self.assertEqual(s1.shape(), unsigned(1))
		self.assertIsInstance(s1.shape(), Shape)
		s2 = Const(-10)[0:2]
		self.assertEqual(s2.shape(), unsigned(2))

	def test_start_end_negative(self):
		c  = Const(0, 8)
		s1 = Slice(c, 0, -1)
		self.assertEqual((s1.start, s1.stop), (0, 7))
		s1 = Slice(c, -4, -1)
		self.assertEqual((s1.start, s1.stop), (4, 7))

	def test_start_end_bool(self):
		c  = Const(0, 8)
		s  = Slice(c, False, True)
		self.assertIs(type(s.start), int)
		self.assertIs(type(s.stop),  int)

	def test_start_end_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Slice start must be an integer, not \'x\'$'
		):
			Slice(0, 'x', 1)
		with self.assertRaisesRegex(
			TypeError,
			r'^Slice stop must be an integer, not \'x\'$'
		):
			Slice(0, 1, 'x')

	def test_start_end_out_of_range(self):
		c = Const(0, 8)
		with self.assertRaisesRegex(
			IndexError,
			r'^Cannot start slice 10 bits into 8-bit value$'
		):
			Slice(c, 10, 12)
		with self.assertRaisesRegex(
			IndexError,
			r'^Cannot stop slice 12 bits into 8-bit value$'
		):
			Slice(c, 0, 12)
		with self.assertRaisesRegex(
			IndexError,
			r'^Slice start 4 must be less than slice stop 2$'
		):
			Slice(c, 4, 2)
		with self.assertRaisesRegex(
			IndexError,
			r'^Cannot start slice -9 bits into 8-bit value$'
		):
			Slice(c, -9, -5)
		with self.assertRaisesRegex(
			IndexError,
			r'^Cannot stop slice 9 bits into 8-bit value$'
		):
			Slice(c, 5, 9)

	def test_repr(self):
		s1 = Const(10)[2]
		self.assertEqual(repr(s1), '(slice (const 4\'d10) 2:3)')


class BitSelectTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.c = Const(0, 8)
		self.s = Signal(range(self.c.width))

	def test_shape(self):
		s1 = self.c.bit_select(self.s, 2)
		self.assertIsInstance(s1, Part)
		self.assertEqual(s1.shape(), unsigned(2))
		self.assertIsInstance(s1.shape(), Shape)
		s2 = self.c.bit_select(self.s, 0)
		self.assertIsInstance(s2, Part)
		self.assertEqual(s2.shape(), unsigned(0))

	def test_stride(self):
		s1 = self.c.bit_select(self.s, 2)
		self.assertIsInstance(s1, Part)
		self.assertEqual(s1.stride, 1)

	def test_const(self):
		s1 = self.c.bit_select(1, 2)
		self.assertIsInstance(s1, Slice)
		self.assertRepr(s1, '(slice (const 8\'d0) 1:3)')

	def test_width_wrong(self):
		with self.assertRaises(TypeError):
			self.c.bit_select(self.s, -1)

	def test_repr(self):
		s = self.c.bit_select(self.s, 2)
		self.assertEqual(repr(s), '(part (const 8\'d0) (sig s) 2 1)')

	def test_offset_wrong(self):
		with self.assertRaisesRegex(TypeError, r'^Part offset must be unsigned$'):
			self.c.bit_select(self.s.as_signed(), 1)

class WordSelectTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.c = Const(0, 8)
		self.s = Signal(range(self.c.width))

	def test_shape(self):
		s1 = self.c.word_select(self.s, 2)
		self.assertIsInstance(s1, Part)
		self.assertEqual(s1.shape(), unsigned(2))
		self.assertIsInstance(s1.shape(), Shape)

	def test_stride(self):
		s1 = self.c.word_select(self.s, 2)
		self.assertIsInstance(s1, Part)
		self.assertEqual(s1.stride, 2)

	def test_const(self):
		s1 = self.c.word_select(1, 2)
		self.assertIsInstance(s1, Slice)
		self.assertRepr(s1, '(slice (const 8\'d0) 2:4)')

	def test_width_wrong(self):
		with self.assertRaises(TypeError):
			self.c.word_select(self.s, 0)
		with self.assertRaises(TypeError):
			self.c.word_select(self.s, -1)

	def test_repr(self):
		s = self.c.word_select(self.s, 2)
		self.assertEqual(repr(s), '(part (const 8\'d0) (sig s) 2 2)')

	def test_offset_wrong(self):
		with self.assertRaisesRegex(TypeError, r'^Part offset must be unsigned$'):
			self.c.word_select(self.s.as_signed(), 1)

class CatTestCase(ToriiTestSuiteCase):
	def test_shape(self):
		c0 = Cat()
		self.assertEqual(c0.shape(), unsigned(0))
		self.assertIsInstance(c0.shape(), Shape)
		c1 = Cat(Const(10))
		self.assertEqual(c1.shape(), unsigned(4))
		c2 = Cat(Const(10), Const(1))
		self.assertEqual(c2.shape(), unsigned(5))
		c3 = Cat(Const(10), Const(1), Const(0))
		self.assertEqual(c3.shape(), unsigned(6))

	def test_repr(self):
		c1 = Cat(Const(10), Const(1))
		self.assertEqual(repr(c1), '(cat (const 4\'d10) (const 1\'d1))')

	def test_cast(self):
		c = Cat(1, 0)
		self.assertEqual(repr(c), '(cat (const 1\'d1) (const 1\'d0))')

	def test_str_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Object \'foo\' cannot be converted to a Torii value$'
		):
			Cat('foo')

	def test_int_01(self):
		with warnings.catch_warnings():
			warnings.filterwarnings(action = 'error', category = SyntaxWarning)
			Cat(0, 1, 1, 0)

	def test_enum_wrong(self):
		class Color(Enum):
			RED  = 1
			BLUE = 2

		with self.assertWarnsRegex(
			SyntaxWarning,
			r'^Argument #1 of \'Cat\(\)\' is an enumerated value <Color\.RED: 1> without '
			r'a defined shape used in a bit vector context; use \'Const\' to specify '
			r'the shape\.$',
		):
			c = Cat(Color.RED, Color.BLUE)
		self.assertEqual(repr(c), '(cat (const 2\'d1) (const 2\'d2))')

	def test_intenum_wrong(self):
		class Color(int, Enum):
			RED  = 1
			BLUE = 2

		with self.assertWarnsRegex(
			SyntaxWarning,
			r'^Argument #1 of \'Cat\(\)\' is an enumerated value <Color\.RED: 1> '
			r'without a defined shape used in a bit vector context; use \'Const\' to specify '
			r'the shape\.$'

		):
			c = Cat(Color.RED, Color.BLUE)
		self.assertEqual(repr(c), '(cat (const 2\'d1) (const 2\'d2))')

	def test_int_wrong(self):
		with self.assertWarnsRegex(
			SyntaxWarning,
			r'^Argument #1 of \'Cat\(\)\' is a bare integer 2 used in bit vector context; '
			r'consider specifying the width explicitly using \'Const\(2, 2\)\' instead$'
		):
			Cat(2)


class ReplTestCase(ToriiTestSuiteCase):
	def test_cast(self):
		r = Repl(0, 3)
		self.assertEqual(repr(r), '(cat (const 1\'d0) (const 1\'d0) (const 1\'d0))')

	def test_int_01(self):
		with warnings.catch_warnings():
			warnings.filterwarnings(action = 'error', category = SyntaxWarning)
			Repl(0, 3)
			Repl(1, 3)

	def test_int_wrong(self):
		with self.assertWarnsRegex(
			SyntaxWarning,
			r'^Value argument of Repl\(\) is a bare integer 2 used in bit vector context; '
			r'consider specifying explicit width using Const\(2, 2\) instead$'
		):
			Repl(2, 3)


class ArrayTestCase(ToriiTestSuiteCase):
	def test_acts_like_array(self):
		a = Array([1, 2, 3])
		self.assertSequenceEqual(a, [1, 2, 3])
		self.assertEqual(a[1], 2)
		a[1] = 4
		self.assertSequenceEqual(a, [1, 4, 3])
		del a[1]
		self.assertSequenceEqual(a, [1, 3])
		a.insert(1, 2)
		self.assertSequenceEqual(a, [1, 2, 3])

	def test_becomes_immutable(self):
		a = Array([1, 2, 3])
		s1 = Signal(range(len(a)))
		s2 = Signal(range(len(a)))
		a[s1]
		a[s2]
		with self.assertRaisesRegex(
			ValueError,
			r'^Array can no longer be mutated after it was indexed with a value at '
		):
			a[1] = 2
		with self.assertRaisesRegex(
			ValueError,
			r'^Array can no longer be mutated after it was indexed with a value at '
		):
			del a[1]
		with self.assertRaisesRegex(
			ValueError,
			r'^Array can no longer be mutated after it was indexed with a value at '
		):
			a.insert(1, 2)

	def test_repr(self):
		a = Array([1, 2, 3])
		self.assertEqual(repr(a), '(array mutable [1, 2, 3])')
		s = Signal(range(len(a)))
		a[s]
		self.assertEqual(repr(a), '(array [1, 2, 3])')


class ArrayProxyTestCase(ToriiTestSuiteCase):
	def test_index_shape(self):
		m = Array(Array(x * y for y in range(1, 4)) for x in range(1, 4))
		a = Signal(range(3))
		b = Signal(range(3))
		v = m[a][b]
		self.assertEqual(v.shape(), unsigned(4))

	def test_attr_shape(self):
		from collections import namedtuple
		pair = namedtuple('pair', ('p', 'n'))
		a = Array(pair(i, -i) for i in range(10))
		s = Signal(range(len(a)))
		v = a[s]
		self.assertEqual(v.p.shape(), unsigned(4))
		self.assertEqual(v.n.shape(), signed(5))

	def test_attr_shape_signed(self):
		# [unsigned(1), unsigned(1)] → unsigned(1)
		a1 = Array([1, 1])
		v1 = a1[Const(0)]
		self.assertEqual(v1.shape(), unsigned(1))
		# [signed(1), signed(1)] → signed(1)
		a2 = Array([-1, -1])
		v2 = a2[Const(0)]
		self.assertEqual(v2.shape(), signed(1))
		# [unsigned(1), signed(2)] → signed(2)
		a3 = Array([1, -2])
		v3 = a3[Const(0)]
		self.assertEqual(v3.shape(), signed(2))
		# [unsigned(1), signed(1)] → signed(2); 1st operand padded with sign bit!
		a4 = Array([1, -1])
		v4 = a4[Const(0)]
		self.assertEqual(v4.shape(), signed(2))
		# [unsigned(2), signed(1)] → signed(3); 1st operand padded with sign bit!
		a5 = Array([1, -1])
		v5 = a5[Const(0)]
		self.assertEqual(v5.shape(), signed(2))

	def test_repr(self):
		a = Array([1, 2, 3])
		s = Signal(range(3))
		v = a[s]
		self.assertEqual(repr(v), '(proxy (array [1, 2, 3]) (sig s))')


class SignalTestCase(ToriiTestSuiteCase):
	def test_shape(self):
		s1 = Signal()
		self.assertEqual(s1.shape(), unsigned(1))
		self.assertIsInstance(s1.shape(), Shape)
		s2 = Signal(2)
		self.assertEqual(s2.shape(), unsigned(2))
		s3 = Signal(unsigned(2))
		self.assertEqual(s3.shape(), unsigned(2))
		s4 = Signal(signed(2))
		self.assertEqual(s4.shape(), signed(2))
		s5 = Signal(0)
		self.assertEqual(s5.shape(), unsigned(0))
		s6 = Signal(range(16))
		self.assertEqual(s6.shape(), unsigned(4))
		s7 = Signal(range(4, 16))
		self.assertEqual(s7.shape(), unsigned(4))
		s8 = Signal(range(-4, 16))
		self.assertEqual(s8.shape(), signed(5))
		s9 = Signal(range(-20, 16))
		self.assertEqual(s9.shape(), signed(6))
		with warnings.catch_warnings():
			warnings.filterwarnings(action = 'ignore', category = SyntaxWarning)
			s10 = Signal(range(0))
			self.assertEqual(s10.shape(), unsigned(0))
		s11 = Signal(range(1))
		self.assertEqual(s11.shape(), unsigned(1))

	def test_shape_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Width of an unsigned value must be zero or a positive integer, not -10$'
		):
			Signal(-10)

	def test_name(self):
		s1 = Signal()
		self.assertEqual(s1.name, 's1')
		s2 = Signal(name = 'sig')
		self.assertEqual(s2.name, 'sig')

	def test_reset(self):
		s1 = Signal(4, reset = 0b111, reset_less = True)
		self.assertEqual(s1.reset, 0b111)
		self.assertEqual(s1.reset_less, True)

	def test_reset_enum(self):
		s1 = Signal(2, reset = UnsignedEnum.BAR)
		self.assertEqual(s1.reset, 2)
		with self.assertRaisesRegex(
			TypeError,
			r'^Reset value must be a constant-castable expression, '
			r'not <StringEnum\.FOO: \'a\'>$'
		):
			Signal(1, reset = StringEnum.FOO)

	def test_reset_signed_mismatch(self):
		with self.assertWarnsRegex(
			SyntaxWarning,
			r'^Reset value -2 is signed, but the signal shape is unsigned\(2\)$'
		):
			Signal(unsigned(2), reset = -2)

	def test_reset_wrong_too_wide(self):
		with self.assertWarnsRegex(
			SyntaxWarning,
			r'^^Reset value 2 will be truncated to the signal shape unsigned\(1\)$'
		):
			Signal(unsigned(1), reset = 2)
		with self.assertWarnsRegex(
			SyntaxWarning,
			r'^Reset value 1 will be truncated to the signal shape signed\(1\)$'
		):
			Signal(signed(1), reset = 1)
		with self.assertWarnsRegex(
			SyntaxWarning,
			r'^Reset value -2 will be truncated to the signal shape signed\(1\)$'
		):
			Signal(signed(1), reset = -2)

	def test_reset_wrong_fencepost(self):
		with self.assertWarnsRegex(
			SyntaxWarning,
			r'^Reset value 10 equals the non-inclusive end of the signal shape '
			r'range\(0, 10\); this is likely an off-by-one error$'
		):
			Signal(range(0, 10), reset=10)


	def test_attrs(self):
		s1 = Signal()
		self.assertEqual(s1.attrs, {})
		s2 = Signal(attrs = {'no_retiming': True})
		self.assertEqual(s2.attrs, {'no_retiming': True})

	def test_repr(self):
		s1 = Signal()
		self.assertEqual(repr(s1), '(sig s1)')

	def test_like(self):
		s1 = Signal.like(Signal(4))
		self.assertEqual(s1.shape(), unsigned(4))
		s2 = Signal.like(Signal(range(-15, 1)))
		self.assertEqual(s2.shape(), signed(5))
		s3 = Signal.like(Signal(4, reset = 0b111, reset_less = True))
		self.assertEqual(s3.reset, 0b111)
		self.assertEqual(s3.reset_less, True)
		s4 = Signal.like(Signal(attrs = {'no_retiming': True}))
		self.assertEqual(s4.attrs, {'no_retiming': True})
		s5 = Signal.like(Signal(decoder = str))
		self.assertEqual(s5.decoder, str)
		s6 = Signal.like(10)
		self.assertEqual(s6.shape(), unsigned(4))
		s7 = [Signal.like(Signal(4))][0]
		self.assertEqual(s7.name, '$like')
		s8 = Signal.like(s1, name_suffix = '_ff')
		self.assertEqual(s8.name, 's1_ff')

	def test_decoder(self):
		class Color(Enum):
			RED  = 1
			BLUE = 2
		s = Signal(decoder = Color)
		self.assertEqual(s.decoder(1), 'RED/1')
		self.assertEqual(s.decoder(3), '3')

	def test_enum(self):
		s1 = Signal(UnsignedEnum)
		self.assertEqual(s1.shape(), unsigned(2))
		s2 = Signal(SignedEnum)
		self.assertEqual(s2.shape(), signed(2))
		self.assertEqual(s2.decoder(SignedEnum.FOO), 'FOO/-1')


class ClockSignalTestCase(ToriiTestSuiteCase):
	def test_domain(self):
		s1 = ClockSignal()
		self.assertEqual(s1.domain, 'sync')
		s2 = ClockSignal('pix')
		self.assertEqual(s2.domain, 'pix')

		with self.assertRaisesRegex(
			TypeError,
			r'^Clock domain name must be a string, not 1$'
		):
			ClockSignal(1)

	def test_shape(self):
		s1 = ClockSignal()
		self.assertEqual(s1.shape(), unsigned(1))
		self.assertIsInstance(s1.shape(), Shape)

	def test_repr(self):
		s1 = ClockSignal()
		self.assertEqual(repr(s1), '(clk sync)')

	def test_wrong_name_comb(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Domain \'comb\' does not have a clock$'
		):
			ClockSignal('comb')


class ResetSignalTestCase(ToriiTestSuiteCase):
	def test_domain(self):
		s1 = ResetSignal()
		self.assertEqual(s1.domain, 'sync')
		s2 = ResetSignal('pix')
		self.assertEqual(s2.domain, 'pix')

		with self.assertRaisesRegex(
			TypeError,
			r'^Clock domain name must be a string, not 1$'
		):
			ResetSignal(1)

	def test_shape(self):
		s1 = ResetSignal()
		self.assertEqual(s1.shape(), unsigned(1))
		self.assertIsInstance(s1.shape(), Shape)

	def test_repr(self):
		s1 = ResetSignal()
		self.assertEqual(repr(s1), '(rst sync)')

	def test_wrong_name_comb(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Domain \'comb\' does not have a reset$'
		):
			ResetSignal('comb')

class MockValueCastable(ValueCastable):
	def __init__(self, dest):
		self.dest = dest

	def shape(self):
		return Value.cast(self.dest).shape()

	@ValueCastable.lowermethod
	def as_value(self):
		return self.dest


class MockValueCastableChanges(ValueCastable):
	def __init__(self, width = 0):
		self.width = width

	def shape(self):
		return unsigned(self.width)

	@ValueCastable.lowermethod
	def as_value(self):
		return Signal(self.width)

class MockValueCastableCustomGetattr(ValueCastable):
	def __init__(self):
		pass

	def shape(self):
		raise AssertionError(False)

	@ValueCastable.lowermethod
	def as_value(self):
		return Const(0)

	def __getattr__(self, attr):
		assert False


class ValueCastableTestCase(ToriiTestSuiteCase):
	def test_not_decorated(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Class \'MockValueCastableNotDecorated\' deriving from `ValueCastable` must '
			r'decorate the `as_value` method with the `ValueCastable.lowermethod` decorator$'
		):
			class MockValueCastableNotDecorated(ValueCastable):
				def __init__(self):
					pass

				def shape(self):
					pass

				def as_value(self):
					return Signal()


	def test_no_override(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Class \'MockValueCastableNoOverrideAsValue\' deriving from `ValueCastable` must '
			r'override the `as_value` method$'
		):
			class MockValueCastableNoOverrideAsValue(ValueCastable):
				def __init__(self):
					pass

		with self.assertRaisesRegex(
			TypeError,
			r'^Class \'MockValueCastableNoOverrideShape\' deriving from `ValueCastable` must '
			r'override the `shape` method$'
		):
			class MockValueCastableNoOverrideShape(ValueCastable):
				def __init__(self):
					pass

				def as_value(self):
					return Signal()

	def test_memoized(self):
		vc = MockValueCastableChanges(1)
		sig1 = vc.as_value()
		vc.width = 2
		sig2 = vc.as_value()
		self.assertIs(sig1, sig2)
		vc.width = 3
		sig3 = Value.cast(vc)
		self.assertIs(sig1, sig3)

	def test_custom_getattr(self):
		vc = MockValueCastableCustomGetattr()
		vc.as_value() # shouldn't call __getattr__

	def test_recurse_bad(self):
		vc = MockValueCastable(None)
		vc.dest = vc
		with self.assertRaisesRegex(
			RecursionError,
			r'^Value-castable object <.+> casts to itself$'
		):
			Value.cast(vc)

	def test_recurse(self):
		vc = MockValueCastable(MockValueCastable(Signal()))
		self.assertIsInstance(Value.cast(vc), Signal)


class SampleTestCase(ToriiTestSuiteCase):
	def test_const(self):
		s = Sample(1, 1, 'sync')
		self.assertEqual(s.shape(), unsigned(1))

	def test_signal(self):
		s1 = Sample(Signal(2), 1, 'sync')
		self.assertEqual(s1.shape(), unsigned(2))

		Sample(ClockSignal(), 1, 'sync')
		Sample(ResetSignal(), 1, 'sync')

	def test_wrong_value_operator(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Sampled value must be a signal or a constant, not '
				r'\(\+ \(sig \$signal\) \(const 1\'d1\)\)$'
			)
		):
			Sample(Signal() + 1, 1, 'sync')

	def test_wrong_clocks_neg(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Cannot sample a value 1 cycles in the future$'
		):
			Sample(Signal(), -1, 'sync')

	def test_wrong_domain(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Domain name must be a string or None, not 0$'
		):
			Sample(Signal(), 1, 0)


class InitialTestCase(ToriiTestSuiteCase):
	def test_initial(self):
		i = Initial()
		self.assertEqual(i.shape(), unsigned(1))


class SwitchTestCase(ToriiTestSuiteCase):
	def test_default_case(self):
		s = Switch(Const(0), {None: []})
		self.assertEqual(s.cases, {(): []})

	def test_int_case(self):
		s = Switch(Const(0, 8), {10: []})
		self.assertEqual(s.cases, {('00001010',): []})

	def test_int_neg_case(self):
		s = Switch(Const(0, 8), {-10: []})
		self.assertEqual(s.cases, {('11110110',): []})

	def test_enum_case(self):
		s = Switch(Const(0, UnsignedEnum), {UnsignedEnum.FOO: []})
		self.assertEqual(s.cases, {('01',): []})

	def test_str_case(self):
		s = Switch(Const(0, 8), {'0000 11\t01': []})
		self.assertEqual(s.cases, {('00001101',): []})

	def test_two_cases(self):
		s = Switch(Const(0, 8), {('00001111', 123): []})
		self.assertEqual(s.cases, {('00001111', '01111011'): []})



# SPDX-License-Identifier: BSD-2-Clause

from collections     import OrderedDict

from torii.build.dsl import *

from ..utils         import ToriiTestSuiteCase


class PinsTestCase(ToriiTestSuiteCase):
	def test_basic(self):
		p = Pins('A0 A1 A2')
		self.assertEqual(repr(p), '(pins io A0 A1 A2)')
		self.assertEqual(len(p.names), 3)
		self.assertEqual(p.dir, 'io')
		self.assertEqual(p.invert, False)
		self.assertEqual(list(p), ['A0', 'A1', 'A2'])

	def test_invert(self):
		p = PinsN('A0')
		self.assertEqual(repr(p), '(pins-n io A0)')
		self.assertEqual(p.invert, True)

	def test_invert_arg(self):
		p = Pins('A0', invert = True)
		self.assertEqual(p.invert, True)

	def test_conn(self):
		p = Pins('0 1 2', conn = ('pmod', 0))
		self.assertEqual(list(p), ['pmod_0:0', 'pmod_0:1', 'pmod_0:2'])
		p = Pins('0 1 2', conn = ('pmod', 'a'))
		self.assertEqual(list(p), ['pmod_a:0', 'pmod_a:1', 'pmod_a:2'])

	def test_map_names(self):
		p = Pins('0 1 2', conn = ('pmod', 0))
		mapping = {
			'pmod_0:0': 'A0',
			'pmod_0:1': 'A1',
			'pmod_0:2': 'A2',
		}
		self.assertEqual(p.map_names(mapping, p), ['A0', 'A1', 'A2'])

	def test_map_names_recur(self):
		p = Pins('0', conn = ('pmod', 0))
		mapping = {
			'pmod_0:0': 'ext_0:1',
			'ext_0:1':  'A1',
		}
		self.assertEqual(p.map_names(mapping, p), ['A1'])

	def test_wrong_names(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Names must be a whitespace-separated string, not \[\'A0\', \'A1\', \'A2\'\]$'
		):
			Pins(['A0', 'A1', 'A2'])

	def test_wrong_dir(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Direction must be one of \'i\', \'o\', \'oe\', or \'io\', not \'wrong\'$'
		):
			Pins('A0 A1', dir = 'wrong')

	def test_wrong_conn(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Connector must be None or a pair of string \(connector name\) and '
				r'integer\/string \(connector number\), not \(\'foo\', None\)$'
			)
		):
			Pins('A0 A1', conn = ('foo', None))

	def test_wrong_map_names(self):
		p = Pins('0 1 2', conn = ('pmod', 0))
		mapping = {
			'pmod_0:0': 'A0',
		}
		with self.assertRaisesRegex(
			NameError, (
				r'^Resource \(pins io pmod_0:0 pmod_0:1 pmod_0:2\) refers to nonexistent '
				r'connector pin pmod_0:1$'
			)
		):
			p.map_names(mapping, p)

	def test_wrong_assert_width(self):
		with self.assertRaisesRegex(
			AssertionError,
			r'^3 names are specified \(0 1 2\), but 4 names are expected$'
		):
			Pins('0 1 2', assert_width = 4)


class DiffPairsTestCase(ToriiTestSuiteCase):
	def test_basic(self):
		dp = DiffPairs(p = 'A0 A1', n = 'B0 B1')
		self.assertEqual(repr(dp), '(diffpairs io (p A0 A1) (n B0 B1))')
		self.assertEqual(dp.p.names, ['A0', 'A1'])
		self.assertEqual(dp.n.names, ['B0', 'B1'])
		self.assertEqual(dp.dir, 'io')
		self.assertEqual(list(dp), [('A0', 'B0'), ('A1', 'B1')])

	def test_invert(self):
		dp = DiffPairsN(p = 'A0', n = 'B0')
		self.assertEqual(repr(dp), '(diffpairs-n io (p A0) (n B0))')
		self.assertEqual(dp.p.names, ['A0'])
		self.assertEqual(dp.n.names, ['B0'])
		self.assertEqual(dp.invert, True)

	def test_invert_arg(self):
		dp = DiffPairs(p = 'A0', n = 'B0', invert = True)
		self.assertEqual(dp.invert, True)

	def test_conn(self):
		dp = DiffPairs(p = '0 1 2', n = '3 4 5', conn = ('pmod', 0))
		self.assertEqual(list(dp), [
			('pmod_0:0', 'pmod_0:3'),
			('pmod_0:1', 'pmod_0:4'),
			('pmod_0:2', 'pmod_0:5'),
		])

	def test_dir(self):
		dp = DiffPairs('A0', 'B0', dir = 'o')
		self.assertEqual(dp.dir, 'o')
		self.assertEqual(dp.p.dir, 'o')
		self.assertEqual(dp.n.dir, 'o')

	def test_wrong_width(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Positive and negative pins must have the same width, but \(pins io A0\) '
				r'and \(pins io B0 B1\) do not$'
			)
		):
			DiffPairs('A0', 'B0 B1')

	def test_wrong_assert_width(self):
		with self.assertRaisesRegex(
			AssertionError,
			r'^3 names are specified \(0 1 2\), but 4 names are expected$'
		):
			DiffPairs('0 1 2', '3 4 5', assert_width = 4)


class AttrsTestCase(ToriiTestSuiteCase):
	def test_basic(self):
		a = Attrs(IO_STANDARD = 'LVCMOS33', PULLUP = 1)
		self.assertEqual(a['IO_STANDARD'], 'LVCMOS33')
		self.assertEqual(repr(a), '(attrs IO_STANDARD=\'LVCMOS33\' PULLUP=1)')

	def test_remove(self):
		a = Attrs(FOO = None)
		self.assertEqual(a['FOO'], None)
		self.assertEqual(repr(a), '(attrs !FOO)')

	def test_callable(self):
		def fn(self):
			return 'FOO'

		a = Attrs(FOO = fn)
		self.assertEqual(a['FOO'], fn)
		self.assertEqual(repr(a), '(attrs FOO={!r})'.format(fn))

	def test_wrong_value(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Value of attribute FOO must be None, int, str, or callable, not 1\.0$'
		):
			Attrs(FOO = 1.0)


class ClockTestCase(ToriiTestSuiteCase):
	def test_basic(self):
		c = Clock(1_000_000)
		self.assertEqual(c.frequency, 1e6)
		self.assertEqual(c.period, 1e-6)
		self.assertEqual(repr(c), '(clock 1000000.0)')


class SubsignalTestCase(ToriiTestSuiteCase):
	def test_basic_pins(self):
		s = Subsignal('a', Pins('A0'), Attrs(IOSTANDARD = 'LVCMOS33'))
		self.assertEqual(
			repr(s),
			'(subsignal a (pins io A0) (attrs IOSTANDARD=\'LVCMOS33\'))'
		)

	def test_basic_diffpairs(self):
		s = Subsignal('a', DiffPairs('A0', 'B0'))
		self.assertEqual(
			repr(s),
			'(subsignal a (diffpairs io (p A0) (n B0)))'
		)

	def test_basic_subsignals(self):
		s = Subsignal('a',
			Subsignal('b', Pins('A0')),
			Subsignal('c', Pins('A1'))
		)
		self.assertEqual(
			repr(s),
			'(subsignal a (subsignal b (pins io A0)) '
			'(subsignal c (pins io A1)))'
		)

	def test_attrs(self):
		s = Subsignal('a',
			Subsignal('b', Pins('A0')),
			Subsignal('c', Pins('A0'), Attrs(SLEW = 'FAST')),
			Attrs(IOSTANDARD = 'LVCMOS33')
		)
		self.assertEqual(s.attrs, {'IOSTANDARD': 'LVCMOS33'})
		self.assertEqual(s.ios[0].attrs, {})
		self.assertEqual(s.ios[1].attrs, {'SLEW': 'FAST'})

	def test_attrs_many(self):
		s = Subsignal('a', Pins('A0'), Attrs(SLEW = 'FAST'), Attrs(PULLUP = '1'))
		self.assertEqual(s.attrs, {'SLEW': 'FAST', 'PULLUP': '1'})

	def test_clock(self):
		s = Subsignal('a', Pins('A0'), Clock(1e6))
		self.assertEqual(s.clock.frequency, 1e6)

	def test_wrong_empty_io(self):
		with self.assertRaisesRegex(ValueError, r'^Missing I\/O constraints$'):
			Subsignal('a')

	def test_wrong_io(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Constraint must be one of Pins, DiffPairs, Subsignal, Attrs, or Clock, '
				r'not \'wrong\'$'
			)
		):
			Subsignal('a', 'wrong')

	def test_wrong_pins(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Pins and DiffPairs are incompatible with other location or subsignal '
				r'constraints, but \(pins io A1\) appears after \(pins io A0\)$'
			)
		):
			Subsignal('a', Pins('A0'), Pins('A1'))

	def test_wrong_diffpairs(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Pins and DiffPairs are incompatible with other location or subsignal '
				r'constraints, but \(pins io A1\) appears after \(diffpairs io \(p A0\) \(n B0\)\)$'
			)
		):
			Subsignal('a', DiffPairs('A0', 'B0'), Pins('A1'))

	def test_wrong_subsignals(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Pins and DiffPairs are incompatible with other location or subsignal '
				r'constraints, but \(pins io B0\) appears after \(subsignal b \(pins io A0\)\)$'
			)
		):
			Subsignal('a', Subsignal('b', Pins('A0')), Pins('B0'))

	def test_wrong_clock(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Clock constraint can only be applied to Pins or DiffPairs, not '
				r'\(subsignal b \(pins io A0\)\)$'
			)
		):
			Subsignal('a', Subsignal('b', Pins('A0')), Clock(1e6))

	def test_wrong_clock_many(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Clock constraint can be applied only once$'
		):
			Subsignal('a', Pins('A0'), Clock(1e6), Clock(1e7))


class ResourceTestCase(ToriiTestSuiteCase):
	def test_basic(self):
		r = Resource('serial', 0,
			Subsignal('tx', Pins('A0', dir = 'o')),
			Subsignal('rx', Pins('A1', dir = 'i')),
			Attrs(IOSTANDARD = 'LVCMOS33')
		)
		self.assertEqual(
			repr(r),
			'(resource serial 0'
			' (subsignal tx (pins o A0))'
			' (subsignal rx (pins i A1))'
			' (attrs IOSTANDARD=\'LVCMOS33\'))'
		)

	def test_number_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Resource number must be an integer, not \(pins o 1\)$'
		):
			# number omitted by accident
			Resource('led', Pins('1', dir = 'o'))

	def test_family(self):
		ios = [Subsignal('clk', Pins('A0', dir = 'o'))]
		r1  = Resource.family(0, default_name = 'spi', ios = ios)
		r2  = Resource.family('spi_flash', 0, default_name = 'spi', ios = ios)
		r3  = Resource.family('spi_flash', 0, default_name = 'spi', ios = ios, name_suffix = '4x')
		r4  = Resource.family(0, default_name = 'spi', ios = ios, name_suffix = '2x')
		self.assertEqual(r1.name, 'spi')
		self.assertEqual(r1.ios, ios)
		self.assertEqual(r2.name, 'spi_flash')
		self.assertEqual(r2.ios, ios)
		self.assertEqual(r3.name, 'spi_flash_4x')
		self.assertEqual(r3.ios, ios)
		self.assertEqual(r4.name, 'spi_2x')
		self.assertEqual(r4.ios, ios)


class ConnectorTestCase(ToriiTestSuiteCase):
	def test_string(self):
		c = Connector('pmod', 0, 'A0 A1 A2 A3 - - A4 A5 A6 A7 - -')
		self.assertEqual(c.name, 'pmod')
		self.assertEqual(c.number, 0)
		self.assertEqual(c.mapping, OrderedDict([
			('1', 'A0'),
			('2', 'A1'),
			('3', 'A2'),
			('4', 'A3'),
			('7', 'A4'),
			('8', 'A5'),
			('9', 'A6'),
			('10', 'A7'),
		]))
		self.assertEqual(list(c), [
			('pmod_0:1', 'A0'),
			('pmod_0:2', 'A1'),
			('pmod_0:3', 'A2'),
			('pmod_0:4', 'A3'),
			('pmod_0:7', 'A4'),
			('pmod_0:8', 'A5'),
			('pmod_0:9', 'A6'),
			('pmod_0:10', 'A7'),
		])
		self.assertEqual(
			repr(c),
			'(connector pmod 0 1=>A0 2=>A1 3=>A2 4=>A3 7=>A4 8=>A5 9=>A6 10=>A7)'
		)

	def test_dict(self):
		c = Connector('ext', 1, {'DP0': 'A0', 'DP1': 'A1'})
		self.assertEqual(c.name, 'ext')
		self.assertEqual(c.number, 1)
		self.assertEqual(c.mapping, OrderedDict([
			('DP0', 'A0'),
			('DP1', 'A1'),
		]))

	def test_conn(self):
		c = Connector('pmod', 0, '0 1 2 3 - - 4 5 6 7 - -', conn = ('expansion', 0))
		self.assertEqual(c.mapping, OrderedDict([
			('1', 'expansion_0:0'),
			('2', 'expansion_0:1'),
			('3', 'expansion_0:2'),
			('4', 'expansion_0:3'),
			('7', 'expansion_0:4'),
			('8', 'expansion_0:5'),
			('9', 'expansion_0:6'),
			('10', 'expansion_0:7'),
		]))

	def test_str_name(self):
		c = Connector('ext', 'A', '0 1 2')
		self.assertEqual(c.name, 'ext')
		self.assertEqual(c.number, 'A')

	def test_conn_wrong_name(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Connector must be None or a pair of string \(connector name\) and '
				r'integer\/string \(connector number\), not \(\'foo\', None\)$'
			)
		):
			Connector('ext', 'A', '0 1 2', conn = ('foo', None))

	def test_wrong_io(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Connector I\/Os must be a dictionary or a string, not \[\]$'
		):
			Connector('pmod', 0, [])

	def test_wrong_dict_key_value(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Connector pin name must be a string, not 0$'
		):
			Connector('pmod', 0, {0: 'A'})
		with self.assertRaisesRegex(
			TypeError,
			r'^Platform pin name must be a string, not 0$'
		):
			Connector('pmod', 0, {'A': 0})



# SPDX-License-Identifier: BSD-2-Clause

from abc        import abstractmethod
from typing     import Union

from ...build   import Clock, TemplatedPlatform
from ...hdl     import ClockDomain, ClockSignal, Const, Instance, Module, Signal
from ...lib.cdc import ResetSynchronizer

__all__ = (
	'QuicklogicPlatform',
)

class QuicklogicPlatform(TemplatedPlatform):
	'''
	.. rubric:: Symbiflow toolchain

	Required tools:
		* ``symbiflow_synth``
		* ``symbiflow_pack``
		* ``symbiflow_place``
		* ``symbiflow_route``
		* ``symbiflow_write_fasm``
		* ``symbiflow_write_bitstream``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_QLSYMBIFLOW``, if present.

	Available overrides:
		* ``add_constraints``: inserts commands in XDC file.

	'''

	@property
	@abstractmethod
	def device(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def package(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	# Since the QuickLogic version of SymbiFlow toolchain is not upstreamed yet
	# we should distinguish the QuickLogic version from mainline one.
	# QuickLogic toolchain: https://github.com/QuickLogic-Corp/quicklogic-fpga-toolchain/releases
	toolchain = 'QLSymbiflow'

	required_tools = [
		'symbiflow_synth',
		'symbiflow_pack',
		'symbiflow_place',
		'symbiflow_route',
		'symbiflow_write_fasm',
		'symbiflow_write_bitstream',
		'symbiflow_write_openocd',
	]
	file_templates = {
		**TemplatedPlatform.build_script_templates,
		'{{name}}.v': r'''
			/* {{autogenerated}} */
			{{emit_verilog()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.pcf': r'''
			# {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				set_io {{port_name}} {{pin_name}}
			{% endfor %}
		''',
		'{{name}}.xdc': r'''
			# {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				{% for attr_name, attr_value in attrs.items() -%}
					set_property {{attr_name}} {{attr_value}} [get_ports {{port_name|tcl_escape}} }]
				{% endfor %}
			{% endfor %}
			{{get_override("add_constraints")|default("# (add_constraints placeholder)")}}
		''',
		'{{name}}.sdc': r'''
			# {{autogenerated}}
			{% for net_signal, port_signal, frequency in platform.iter_clock_constraints() -%}
				{% if port_signal is not none -%}
					create_clock -period {{100000000/frequency}} {{port_signal.name|ascii_escape}}
				{% endif %}
			{% endfor %}
		'''
	}
	command_templates = [
		r'''
		{{invoke_tool("symbiflow_synth")}}
			-t {{name}}
			-v {% for file in platform.iter_files(".v", ".sv", ".vhd", ".vhdl") -%} {{file}} {% endfor %} {{name}}.v
			-d {{platform.device}}
			-p {{name}}.pcf
			-P {{platform.package}}
			-x {{name}}.xdc
		''',
		r'''
		{{invoke_tool("symbiflow_pack")}}
			-e {{name}}.eblif
			-d {{platform.device}}
			-s {{name}}.sdc
		''',
		r'''
		{{invoke_tool("symbiflow_place")}}
			-e {{name}}.eblif
			-d {{platform.device}}
			-p {{name}}.pcf
			-n {{name}}.net
			-P {{platform.package}}
			-s {{name}}.sdc
		''',
		r'''
		{{invoke_tool("symbiflow_route")}}
			-e {{name}}.eblif
			-d {{platform.device}}
			-s {{name}}.sdc
		''',
		r'''
		{{invoke_tool("symbiflow_write_fasm")}}
			-e {{name}}.eblif
			-d {{platform.device}}
			-s {{name}}.sdc
		''',
		r'''
		{{invoke_tool("symbiflow_write_bitstream")}}
			-f {{name}}.fasm
			-d {{platform.device}}
			-P {{platform.package}}
			-b {{name}}.bit
		''',
		# This should be `invoke_tool("symbiflow_write_openocd")`, but isn't because of a bug in
		# the QLSymbiflow v1.3.0 toolchain release.
		r'''
		python3 -m quicklogic_fasm.bitstream_to_openocd
			{{name}}.bit
			{{name}}.openocd
			--osc-freq {{platform.osc_freq}}
			--fpga-clk-divider {{platform.osc_div}}
		''',
	]

	# Common logic

	@property
	def default_clk_constraint(self) -> Clock:
		if self.default_clk == 'sys_clk0':
			return Clock(self.osc_freq / self.osc_div)
		return super().default_clk_constraint

	def add_clock_constraint(self, clock: Signal, frequency: Union[int, float]) -> None:
		super().add_clock_constraint(clock, frequency)
		clock.attrs['keep'] = 'TRUE'

	def create_missing_domain(self, name: str) -> Module:
		if name == 'sync' and self.default_clk is not None:
			m = Module()
			if self.default_clk == 'sys_clk0':
				if not hasattr(self, 'osc_div'):
					raise ValueError('OSC divider (osc_div) must be an integer between 2 and 512')
				if not isinstance(self.osc_div, int) or self.osc_div < 2 or self.osc_div > 512:
					raise ValueError(
						f'OSC divider (osc_div) must be an integer between 2 and 512, not {self.osc_div!r}'
					)
				if not hasattr(self, 'osc_freq'):
					raise ValueError('OSC frequency (osc_freq) must be an integer between 2100000 and 80000000')
				if not isinstance(self.osc_freq, int) or self.osc_freq < 2100000 or self.osc_freq > 80000000:
					raise ValueError(
						f'OSC frequency (osc_freq) must be an integer between 2100000 and 80000000, not {self.osc_freq!r}'
					)
				clk_i = Signal()
				sys_clk0 = Signal()
				m.submodules += Instance(
					'qlal4s3b_cell_macro',
					o_Sys_Clk0 = sys_clk0
				)
				m.submodules += Instance(
					'gclkbuff',
					o_A = sys_clk0,
					o_Z = clk_i
				)
			else:
				clk_i = self.request(self.default_clk).i

			if self.default_rst is not None:
				rst_i = self.request(self.default_rst).i
			else:
				rst_i = Const(0)

			m.domains += ClockDomain('sync')
			m.d.comb += ClockSignal('sync').eq(clk_i)
			m.submodules.reset_sync = ResetSynchronizer(rst_i, domain = 'sync')
			return m



# SPDX-License-Identifier: BSD-2-Clause

import warnings
from abc               import ABCMeta, abstractmethod
from collections       import OrderedDict, defaultdict
from functools         import reduce
from typing            import Any, Optional, Union

from ..util            import flatten
from ..util.decorators import memoize
from ._unused          import MustUse, UnusedMustUse
from .ast              import (
	ClockSignal, ResetSignal, Signal, SignalDict, SignalSet, Statement,
	Value
)
from .cd               import ClockDomain, DomainError

__all__ = (
	'DriverConflict',
	'Elaboratable',
	'Fragment',
	'Instance',
	'UnusedElaboratable',
)


class UnusedElaboratable(UnusedMustUse):
	# The warning is initially silenced. If everything that has been constructed remains unused,
	# it means the application likely crashed (with an exception, or in another way that does not
	# call `sys.excepthook`), and it's not necessary to show any warnings.
	# Once elaboration starts, the warning is enabled.
	_MustUse__silence = True


class Elaboratable(MustUse, metaclass = ABCMeta):
	_MustUse__warning = UnusedElaboratable

	@abstractmethod
	def elaborate(self, platform):
		''' '''
		raise NotImplementedError('Elaboratables must implement the \'elaborate\' method')


class DriverConflict(UserWarning):
	pass


class Fragment:
	@staticmethod
	def get(
		obj: Union['Fragment', Elaboratable, object] , platform: Optional[Any]
	) -> 'Fragment':
		code = None
		while True:
			if isinstance(obj, Fragment):
				return obj
			elif isinstance(obj, Elaboratable):
				code = obj.elaborate.__code__
				UnusedElaboratable._MustUse__silence = False
				obj._MustUse__used = True
				new_obj = obj.elaborate(platform)
			else:
				raise AttributeError(f'Object {obj!r} cannot be elaborated')
			if new_obj is obj:
				raise RecursionError(f'Object {obj!r} elaborates to itself')
			if new_obj is None and code is not None:
				warnings.warn_explicit(
					'.elaborate() returned None; missing return statement?',
					category = UserWarning,
					filename = code.co_filename,
					lineno = code.co_firstlineno)
			obj = new_obj

	def __init__(self):
		self.ports = SignalDict()
		self.drivers = OrderedDict()
		self.statements = []
		self.domains = OrderedDict()
		self.subfragments = []
		self.attrs = OrderedDict()
		self.generated = OrderedDict()
		self.flatten = False

	def add_ports(self, *ports, dir):
		assert dir in ('i', 'o', 'io')
		for port in flatten(ports):
			self.ports[port] = dir

	def iter_ports(self, dir = None):
		if dir is None:
			yield from self.ports
		else:
			for port, port_dir in self.ports.items():
				if port_dir == dir:
					yield port

	def add_driver(self, signal, domain = None):
		if domain not in self.drivers:
			self.drivers[domain] = SignalSet()
		self.drivers[domain].add(signal)

	def iter_drivers(self):
		for domain, signals in self.drivers.items():
			for signal in signals:
				yield domain, signal

	def iter_comb(self):
		if None in self.drivers:
			yield from self.drivers[None]

	def iter_sync(self):
		for domain, signals in self.drivers.items():
			if domain is None:
				continue
			for signal in signals:
				yield domain, signal

	def iter_signals(self):
		signals = SignalSet()
		signals |= self.ports.keys()
		for domain, domain_signals in self.drivers.items():
			if domain is not None:
				cd = self.domains[domain]
				signals.add(cd.clk)
				if cd.rst is not None:
					signals.add(cd.rst)
			signals |= domain_signals
		return signals

	def add_domains(self, *domains):
		for domain in flatten(domains):
			assert isinstance(domain, ClockDomain)
			assert domain.name not in self.domains
			self.domains[domain.name] = domain

	def iter_domains(self):
		yield from self.domains

	def add_statements(self, *stmts):
		for stmt in Statement.cast(stmts):
			stmt._MustUse__used = True
			self.statements.append(stmt)

	def add_subfragment(self, subfragment, name = None):
		assert isinstance(subfragment, Fragment)
		self.subfragments.append((subfragment, name))

	def find_subfragment(self, name_or_index):
		if isinstance(name_or_index, int):
			if name_or_index < len(self.subfragments):
				subfragment, name = self.subfragments[name_or_index]
				return subfragment
			raise NameError(f'No subfragment at index #{name_or_index}')
		else:
			for subfragment, name in self.subfragments:
				if name == name_or_index:
					return subfragment
			raise NameError(f'No subfragment with name \'{name_or_index}\'')

	def find_generated(self, *path):
		if len(path) > 1:
			path_component, *path = path
			return self.find_subfragment(path_component).find_generated(*path)
		else:
			item, = path
			return self.generated[item]

	def elaborate(self, platform):
		return self

	def _merge_subfragment(self, subfragment):
		# Merge subfragment's everything except clock domains into this fragment.
		# Flattening is done after clock domain propagation, so we can assume the domains
		# are already the same in every involved fragment in the first place.
		self.ports.update(subfragment.ports)
		for domain, signal in subfragment.iter_drivers():
			self.add_driver(signal, domain)
		self.statements += subfragment.statements
		self.subfragments += subfragment.subfragments

		# Remove the merged subfragment.
		found = False
		for i, (check_subfrag, check_name) in enumerate(self.subfragments): # :nobr:
			if subfragment == check_subfrag:
				del self.subfragments[i]
				found = True
				break
		assert found

	def _resolve_hierarchy_conflicts(self, hierarchy = ('top',), mode = 'warn'):
		assert mode in ('silent', 'warn', 'error')

		driver_subfrags = SignalDict()

		def add_subfrag(registry, entity, entry):
			# Because of missing domain insertion, at the point when this code runs, we have
			# a mixture of bound and unbound {Clock,Reset}Signals. Map the bound ones to
			# the actual signals (because the signal itself can be driven as well); but leave
			# the unbound ones as it is, because there's no concrete signal for it yet anyway.
			if isinstance(entity, ClockSignal) and entity.domain in self.domains:
				entity = self.domains[entity.domain].clk
			elif isinstance(entity, ResetSignal) and entity.domain in self.domains:
				entity = self.domains[entity.domain].rst

			if entity not in registry:
				registry[entity] = set()
			registry[entity].add(entry)

		# For each signal driven by this fragment and/or its subfragments, determine which
		# subfragments also drive it.
		for domain, signal in self.iter_drivers():
			add_subfrag(driver_subfrags, signal, (None, hierarchy))

		flatten_subfrags = set()
		for i, (subfrag, name) in enumerate(self.subfragments):
			if name is None:
				name = f'<unnamed #{i}>'
			subfrag_hierarchy = hierarchy + (name,)

			if subfrag.flatten:
				# Always flatten subfragments that explicitly request it.
				flatten_subfrags.add((subfrag, subfrag_hierarchy))

			if isinstance(subfrag, Instance):
				# Never flatten instances.
				continue

			# First, recurse into subfragments and let them detect driver conflicts as well.
			subfrag_drivers = subfrag._resolve_hierarchy_conflicts(
				subfrag_hierarchy, mode
			)

			# Second, classify subfragments by signals they drive.
			for signal in subfrag_drivers:
				add_subfrag(driver_subfrags, signal, (subfrag, subfrag_hierarchy))

		# Find out the set of subfragments that needs to be flattened into this fragment
		# to resolve driver-driver conflicts.
		def flatten_subfrags_if_needed(subfrags):
			if len(subfrags) == 1:
				return []
			flatten_subfrags.update((f, h) for f, h in subfrags if f is not None)
			return list(sorted('.'.join(h) for f, h in subfrags))

		for signal, subfrags in driver_subfrags.items():
			subfrag_names = flatten_subfrags_if_needed(subfrags)
			if not subfrag_names:
				continue

			# While we're at it, show a message.
			message = (f'Signal \'{signal}\' is driven from multiple fragments: {", ".join(subfrag_names)}')
			if mode == 'error':
				raise DriverConflict(message)
			elif mode == 'warn':
				message += '; hierarchy will be flattened'
				warnings.warn_explicit(message, DriverConflict, *signal.src_loc)

		# Flatten hierarchy.
		for subfrag, subfrag_hierarchy in sorted(flatten_subfrags, key = lambda x: x[1]):
			self._merge_subfragment(subfrag)

		# If we flattened anything, we might be in a situation where we have a driver conflict
		# again, e.g. if we had a tree of fragments like A --- B --- C where only fragments
		# A and C were driving a signal S. In that case, since B is not driving S itself,
		# processing B will not result in any flattening, but since B is transitively driving S,
		# processing A will flatten B into it. Afterwards, we have a tree like AB --- C, which
		# has another conflict.
		if any(flatten_subfrags):
			# Try flattening again.
			return self._resolve_hierarchy_conflicts(hierarchy, mode)

		# Nothing was flattened, we're done!
		return SignalSet(driver_subfrags.keys())

	def _propagate_domains_up(self, hierarchy = ('top',)):
		from .xfrm import DomainRenamer

		domain_subfrags = defaultdict(lambda: set())

		# For each domain defined by a subfragment, determine which subfragments define it.
		for i, (subfrag, name) in enumerate(self.subfragments):
			# First, recurse into subfragments and let them propagate domains up as well.
			hier_name = name
			if hier_name is None:
				hier_name = f'<unnamed #{i}>'
			subfrag._propagate_domains_up(hierarchy + (hier_name,))

			# Second, classify subfragments by domains they define.
			for domain_name, domain in subfrag.domains.items():
				if domain.local:
					continue
				domain_subfrags[domain_name].add((subfrag, name, i))

		# For each domain defined by more than one subfragment, rename the domain in each
		# of the subfragments such that they no longer conflict.
		for domain_name, subfrags in domain_subfrags.items():
			if len(subfrags) == 1:
				continue

			names = [n for f, n, i in subfrags]
			if not all(names):
				names = sorted(f'<unnamed #{i}>' if n is None else f'\'{n}\'' for f, n, i in subfrags)
				raise DomainError(
					f'Domain \'{domain_name}\' is defined by subfragments {", ".join(names)} of fragment \'{".".join(hierarchy)}\'; '
					'it is necessary to either rename subfragment domains '
					'explicitly, or give names to subfragments'
				)

			if len(names) != len(set(names)):
				names = sorted(f'#{i}' for f, n, i in subfrags)
				raise DomainError(
					f'Domain \'{domain_name}\' is defined by subfragments {", ".join(names)} of fragment \'{".".join(hierarchy)}\', '
					'some of which have identical names; it is necessary to either '
					'rename subfragment domains explicitly, or give distinct names '
					'to subfragments'
				)

			for subfrag, name, i in subfrags:
				domain_name_map = { domain_name: f'{name}_{domain_name}' }
				self.subfragments[i] = (DomainRenamer(domain_name_map)(subfrag), name)

		# Finally, collect the (now unique) subfragment domains, and merge them into our domains.
		for subfrag, name in self.subfragments:
			for domain_name, domain in subfrag.domains.items():
				if domain.local:
					continue
				self.add_domains(domain)

	def _propagate_domains_down(self):
		# For each domain defined in this fragment, ensure it also exists in all subfragments.
		for subfrag, name in self.subfragments:
			for domain in self.iter_domains():
				if domain in subfrag.domains:
					assert self.domains[domain] is subfrag.domains[domain]
				else:
					subfrag.add_domains(self.domains[domain])

			subfrag._propagate_domains_down()

	def _create_missing_domains(self, missing_domain, *, platform = None):
		from .xfrm import DomainCollector

		collector = DomainCollector()
		collector(self)

		new_domains = []
		for domain_name in collector.used_domains - collector.defined_domains:
			if domain_name is None:
				continue
			value = missing_domain(domain_name)
			if value is None:
				raise DomainError(f'Domain \'{domain_name}\' is used but not defined')
			if type(value) is ClockDomain:
				self.add_domains(value)
				# And expose ports on the newly added clock domain, since it is added directly
				# and there was no chance to add any logic driving it.
				new_domains.append(value)
			else:
				new_fragment = Fragment.get(value, platform = platform)
				if domain_name not in new_fragment.domains:
					defined = new_fragment.domains.keys()
					raise DomainError(
						'Fragment returned by missing domain callback does not define '
						f'requested domain \'{domain_name}\' (defines {", ".join("`{}`".format(n) for n in defined)}).')
				self.add_subfragment(new_fragment, f'cd_{domain_name}')
				self.add_domains(new_fragment.domains.values())
		return new_domains

	def _propagate_domains(self, missing_domain, *, platform = None):
		self._propagate_domains_up()
		self._propagate_domains_down()
		self._resolve_hierarchy_conflicts()
		new_domains = self._create_missing_domains(missing_domain, platform = platform)
		self._propagate_domains_down()
		return new_domains

	def _prepare_use_def_graph(self, parent, level, uses, defs, ios, top):
		def add_uses(*sigs, self = self):
			for sig in flatten(sigs):
				if sig not in uses:
					uses[sig] = set()
				uses[sig].add(self)

		def add_defs(*sigs):
			for sig in flatten(sigs):
				if sig not in defs:
					defs[sig] = self
				else:
					assert defs[sig] is self

		def add_io(*sigs):
			for sig in flatten(sigs):
				if sig not in ios:
					ios[sig] = self
				else:
					assert ios[sig] is self

		# Collect all signals we're driving (on LHS of statements), and signals we're using
		# (on RHS of statements, or in clock domains).
		for stmt in self.statements:
			add_uses(stmt._rhs_signals())
			add_defs(stmt._lhs_signals())

		for domain, _ in self.iter_sync():
			cd = self.domains[domain]
			add_uses(cd.clk)
			if cd.rst is not None:
				add_uses(cd.rst)

		# Repeat for subfragments.
		for subfrag, name in self.subfragments:
			if isinstance(subfrag, Instance):
				for port_name, (value, dir) in subfrag.named_ports.items():
					if dir == 'i':
						# Prioritize defs over uses.
						rhs_without_outputs = value._rhs_signals() - subfrag.iter_ports(dir = 'o')
						subfrag.add_ports(rhs_without_outputs, dir = dir)
						add_uses(value._rhs_signals())
					if dir == 'o':
						subfrag.add_ports(value._lhs_signals(), dir = dir)
						add_defs(value._lhs_signals())
					if dir == 'io':
						subfrag.add_ports(value._lhs_signals(), dir = dir)
						add_io(value._lhs_signals())
			else:
				parent[subfrag] = self
				level[subfrag]  = level[self] + 1

				subfrag._prepare_use_def_graph(parent, level, uses, defs, ios, top)

	def _propagate_ports(self, ports, all_undef_as_ports):
		# Take this fragment graph:
		#
		#    __ B (def: q, use: p r)
		#   /
		#  A (def: p, use: q r)
		#   \
		#    \_ C (def: r, use: p q)
		#
		# We need to consider three cases.
		#   1. Signal p requires an input port in B;
		#   2. Signal r requires an output port in C;
		#   3. Signal r requires an output port in C and an input port in B.
		#
		# Adding these ports can be in general done in three steps for each signal:
		#   1. Find the least common ancestor of all uses and defs.
		#   2. Going upwards from the single def, add output ports.
		#   3. Going upwards from all uses, add input ports.

		parent = { self: None }
		level  = { self: 0 }
		uses   = SignalDict()
		defs   = SignalDict()
		ios    = SignalDict()
		self._prepare_use_def_graph(parent, level, uses, defs, ios, self)

		ports = SignalSet(ports)
		if all_undef_as_ports:
			for sig in uses:
				if sig in defs:
					continue
				ports.add(sig)
		for sig in ports:
			if sig not in uses:
				uses[sig] = set()
			uses[sig].add(self)

		@memoize
		def lca_of(fragu, fragv):
			# Normalize fragu to be deeper than fragv.
			if level[fragu] < level[fragv]:
				fragu, fragv = fragv, fragu
			# Find ancestor of fragu on the same level as fragv.
			for _ in range(level[fragu] - level[fragv]):
				fragu = parent[fragu]
			# If fragv was the ancestor of fragv, we're done.
			if fragu == fragv:
				return fragu
			# Otherwise, they are at the same level but in different branches. Step both fragu
			# and fragv until we find the common ancestor.
			while parent[fragu] != parent[fragv]:
				fragu = parent[fragu]
				fragv = parent[fragv]
			return parent[fragu]

		for sig in uses:
			if sig in defs:
				lca  = reduce(lca_of, uses[sig], defs[sig])
			else:
				lca  = reduce(lca_of, uses[sig])

			for frag in uses[sig]:
				if sig in defs and frag is defs[sig]:
					continue
				while frag != lca:
					frag.add_ports(sig, dir = 'i')
					frag = parent[frag]

			if sig in defs:
				frag = defs[sig]
				while frag != lca:
					frag.add_ports(sig, dir = 'o')
					frag = parent[frag]

		for sig in ios:
			frag = ios[sig]
			while frag is not None:
				frag.add_ports(sig, dir = 'io')
				frag = parent[frag]

		for sig in ports:
			if sig in ios:
				continue
			if sig in defs:
				self.add_ports(sig, dir = 'o')
			else:
				self.add_ports(sig, dir = 'i')

	def prepare(self, ports = None, missing_domain = lambda name: ClockDomain(name)):
		from .xfrm import DomainLowerer, SampleLowerer

		fragment = SampleLowerer()(self)
		new_domains = fragment._propagate_domains(missing_domain)
		fragment = DomainLowerer()(fragment)
		if ports is None:
			fragment._propagate_ports(ports = (), all_undef_as_ports = True)
		else:
			if not isinstance(ports, tuple) and not isinstance(ports, list):
				msg = f'`ports` must be either a list or a tuple, not {ports!r}'
				if isinstance(ports, Value):
					msg += ' (did you mean `ports = (<signal>,)`, rather than `ports = <signal>`?)'
				raise TypeError(msg)
			mapped_ports = []
			# Lower late bound signals like ClockSignal() to ports.
			port_lowerer = DomainLowerer(fragment.domains)
			for port in ports:
				if not isinstance(port, (Signal, ClockSignal, ResetSignal)):
					raise TypeError(f'Only signals may be added as ports, not {port!r}')
				mapped_ports.append(port_lowerer.on_value(port))
			# Add ports for all newly created missing clock domains, since not doing so defeats
			# the purpose of domain auto-creation. (It's possible to refer to these ports before
			# the domain actually exists through late binding, but it's inconvenient.)
			for cd in new_domains:
				mapped_ports.append(cd.clk)
				if cd.rst is not None:
					mapped_ports.append(cd.rst)
			fragment._propagate_ports(ports = mapped_ports, all_undef_as_ports = False)
		return fragment


class Instance(Fragment):
	'''
	Allows for the direct instantiation of external modules, cells, or primitives.

	It accepts the name of the object to instantiate and a collection of keyword arguments
	that define the ports, attributes, and parameters to it.

	It is defined by a prefix followed by the canonical name of the element the value is setting.
	For instance, if you have a cell called ``dff`` with ports named ``CLK``, ``D``, and ``Q`` you
	are able to instantiate it as follows:

	.. code-block:: python

		dff = Instance(
			'dff',
			i_D   = sig_in,
			i_CLK = clk,
			o_Q   = sig_out,
		)

	The meaning of the prefix for the arguments are as follows:

	+---------+----------------------------+
	| Prefix  | Corresponding Type         |
	+=========+============================+
	| ``a_``  | Attribute                  |
	+---------+----------------------------+
	| ``p_``  | Parameter                  |
	+---------+----------------------------+
	| ``i_``  | Input Port/Signal          |
	+---------+----------------------------+
	| ``o_``  | Output Port/Signal         |
	+---------+----------------------------+
	| ``io_`` | Bi-directional Port/Signal |
	+---------+----------------------------+

	Parameters
	----------
	type: str
		The name/type of object to instantiate

	'''

	def __init__(self, type, *args, **kwargs):
		super().__init__()

		self.type        = type
		self.parameters  = OrderedDict()
		self.named_ports = OrderedDict()

		for (kind, name, value) in args:
			if kind == 'a':
				self.attrs[name] = value
			elif kind == 'p':
				self.parameters[name] = value
			elif kind in ('i', 'o', 'io'):
				self.named_ports[name] = (Value.cast(value), kind)
			else:
				raise NameError(
					f'Instance argument {(kind, name, value)!r} should be a tuple (kind, name, value) '
					'where kind is one of \'a\', \'p\', \'i\', \'o\', or \'io\''
				)

		for kw, arg in kwargs.items():
			if kw.startswith('a_'):
				self.attrs[kw[2:]] = arg
			elif kw.startswith('p_'):
				self.parameters[kw[2:]] = arg
			elif kw.startswith('i_'):
				self.named_ports[kw[2:]] = (Value.cast(arg), 'i')
			elif kw.startswith('o_'):
				self.named_ports[kw[2:]] = (Value.cast(arg), 'o')
			elif kw.startswith('io_'):
				self.named_ports[kw[3:]] = (Value.cast(arg), 'io')
			else:
				raise NameError(
					f'Instance keyword argument {kw} = {arg!r} does not start with one of '
					'\'a_\', \'p_\', \'i_\', \'o_\', or \'io_\''
				)



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Module, ClockDomain, Signal, ClockSignal, ResetSignal
from torii.back import verilog

m = Module()
cd_por  = ClockDomain(reset_less = True)
cd_sync = ClockDomain()
m.domains += cd_por, cd_sync

delay = Signal(range(256), reset = 255)
with m.If(delay != 0):
	m.d.por += delay.eq(delay - 1)
m.d.comb += [
	ClockSignal().eq(cd_por.clk),
	ResetSignal().eq(delay != 0),
]

if __name__ == '__main__':
	print(verilog.convert(m, ports = [cd_por.clk]))



# SPDX-License-Identifier: BSD-2-Clause

from .pmod import (
	PmodDualHBridgeType6Resource, PmodGPIOType1Resource,
	PmodHBridgeType5Resource, PmodSPIType2AResource,
	PmodSPIType2Resource, PmodUARTType3Resource,
	PmodUARTType4AResource, PmodUARTType4Resource
)

__all__ = (
	'PmodDualHBridgeType6Resource',
	'PmodGPIOType1Resource',
	'PmodHBridgeType5Resource',
	'PmodSPIType2AResource',
	'PmodSPIType2Resource',
	'PmodUARTType3Resource',
	'PmodUARTType4AResource',
	'PmodUARTType4Resource',
)



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii           import *
from torii.hdl.rec   import *
from torii.lib.io    import *
from torii.build.dsl import *
from torii.build.res import *

from ..utils         import ToriiTestSuiteCase

class ResourceManagerTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.resources = [
			Resource('clk100', 0, DiffPairs('H1', 'H2', dir = 'i'), Clock(100e6)),
			Resource('clk50', 0, Pins('K1'), Clock(50e6)),
			Resource('user_led', 0, Pins('A0', dir = 'o')),
			Resource('i2c', 0,
				Subsignal('scl', Pins('N10', dir = 'o')),
				Subsignal('sda', Pins('N11'))
			)
		]
		self.connectors = [
			Connector('pmod', 0, 'B0 B1 B2 B3 - -'),
		]
		self.cm = ResourceManager(self.resources, self.connectors)

	def test_basic(self):
		self.cm = ResourceManager(self.resources, self.connectors)
		self.assertEqual(self.cm.resources, {
			('clk100',   0): self.resources[0],
			('clk50',    0): self.resources[1],
			('user_led', 0): self.resources[2],
			('i2c',      0): self.resources[3]
		})
		self.assertEqual(self.cm.connectors, {
			('pmod', 0): self.connectors[0],
		})

	def test_add_resources(self):
		new_resources = [
			Resource('user_led', 1, Pins('A1', dir = 'o'))
		]
		self.cm.add_resources(new_resources)
		self.assertEqual(self.cm.resources, {
			('clk100',   0): self.resources[0],
			('clk50',    0): self.resources[1],
			('user_led', 0): self.resources[2],
			('i2c',      0): self.resources[3],
			('user_led', 1): new_resources[0]
		})

	def test_lookup(self):
		r = self.cm.lookup('user_led', 0)
		self.assertIs(r, self.cm.resources['user_led', 0])

	def test_request_basic(self):
		self.cm.lookup('user_led', 0)
		user_led = self.cm.request('user_led', 0)

		self.assertIsInstance(user_led, Pin)
		self.assertEqual(user_led.name, 'user_led_0')
		self.assertEqual(user_led.width, 1)
		self.assertEqual(user_led.dir, 'o')

		ports = list(self.cm.iter_ports())
		self.assertEqual(len(ports), 1)

		self.assertEqual(list(self.cm.iter_port_constraints()), [
			('user_led_0__io', ['A0'], {})
		])

	def test_request_with_dir(self):
		i2c = self.cm.request('i2c', 0, dir = {'sda': 'o'})
		self.assertIsInstance(i2c, Record)
		self.assertIsInstance(i2c.sda, Pin)
		self.assertEqual(i2c.sda.dir, 'o')

	def test_request_tristate(self):
		i2c = self.cm.request('i2c', 0)
		self.assertEqual(i2c.sda.dir, 'io')

		ports = list(self.cm.iter_ports())
		self.assertEqual(len(ports), 2)
		scl, sda = ports
		self.assertEqual(ports[1].name, 'i2c_0__sda__io')
		self.assertEqual(ports[1].width, 1)

		scl_info, sda_info = self.cm.iter_single_ended_pins()
		self.assertIs(scl_info[0], i2c.scl)
		self.assertIs(scl_info[1].io, scl)
		self.assertEqual(scl_info[2], {})
		self.assertEqual(scl_info[3], False)
		self.assertIs(sda_info[0], i2c.sda)
		self.assertIs(sda_info[1].io, sda)

		self.assertEqual(list(self.cm.iter_port_constraints()), [
			('i2c_0__scl__io', ['N10'], {}),
			('i2c_0__sda__io', ['N11'], {})
		])

	def test_request_diffpairs(self):
		clk100 = self.cm.request('clk100', 0)
		self.assertIsInstance(clk100, Pin)
		self.assertEqual(clk100.dir, 'i')
		self.assertEqual(clk100.width, 1)

		ports = list(self.cm.iter_ports())
		self.assertEqual(len(ports), 2)
		p, n = ports
		self.assertEqual(p.name, 'clk100_0__p')
		self.assertEqual(p.width, clk100.width)
		self.assertEqual(n.name, 'clk100_0__n')
		self.assertEqual(n.width, clk100.width)

		clk100_info, = self.cm.iter_differential_pins()
		self.assertIs(clk100_info[0], clk100)
		self.assertIs(clk100_info[1].p, p)
		self.assertIs(clk100_info[1].n, n)
		self.assertEqual(clk100_info[2], {})
		self.assertEqual(clk100_info[3], False)

		self.assertEqual(list(self.cm.iter_port_constraints()), [
			('clk100_0__p', ['H1'], {}),
			('clk100_0__n', ['H2'], {}),
		])

	def test_request_inverted(self):
		new_resources = [
			Resource('cs', 0, PinsN('X0')),
			Resource('clk', 0, DiffPairsN('Y0', 'Y1')),
		]
		self.cm.add_resources(new_resources)

		cs = self.cm.request('cs')
		clk = self.cm.request('clk')
		cs_io, clk_p, clk_n = self.cm.iter_ports()

		cs_info, = self.cm.iter_single_ended_pins()
		self.assertIs(cs_info[0], cs)
		self.assertIs(cs_info[1].io, cs_io)
		self.assertEqual(cs_info[2], {})
		self.assertEqual(cs_info[3], True)

		clk_info, = self.cm.iter_differential_pins()
		self.assertIs(clk_info[0], clk)
		self.assertIs(clk_info[1].p, clk_p)
		self.assertIs(clk_info[1].n, clk_n)
		self.assertEqual(clk_info[2], {})
		self.assertEqual(clk_info[3], True)

	def test_request_raw(self):
		clk50 = self.cm.request('clk50', 0, dir = '-')
		self.assertIsInstance(clk50, Record)
		self.assertIsInstance(clk50.io, Signal)

		ports = list(self.cm.iter_ports())
		self.assertEqual(len(ports), 1)
		self.assertIs(ports[0], clk50.io)

	def test_request_raw_diffpairs(self):
		clk100 = self.cm.request('clk100', 0, dir = '-')
		self.assertIsInstance(clk100, Record)
		self.assertIsInstance(clk100.p, Signal)
		self.assertIsInstance(clk100.n, Signal)

		ports = list(self.cm.iter_ports())
		self.assertEqual(len(ports), 2)
		self.assertIs(ports[0], clk100.p)
		self.assertIs(ports[1], clk100.n)

	def test_request_via_connector(self):
		self.cm.add_resources([
			Resource('spi', 0,
				Subsignal('ss',   Pins('1', conn = ('pmod', 0))),
				Subsignal('clk',  Pins('2', conn = ('pmod', 0))),
				Subsignal('miso', Pins('3', conn = ('pmod', 0))),
				Subsignal('mosi', Pins('4', conn = ('pmod', 0))),
			)
		])
		self.cm.request('spi', 0)
		self.assertEqual(list(self.cm.iter_port_constraints()), [
			('spi_0__ss__io',   ['B0'], {}),
			('spi_0__clk__io',  ['B1'], {}),
			('spi_0__miso__io', ['B2'], {}),
			('spi_0__mosi__io', ['B3'], {}),
		])

	def test_request_via_nested_connector(self):
		new_connectors = [
			Connector('pmod_extension', 0, '1 2 3 4 - -', conn = ('pmod', 0)),
		]
		self.cm.add_connectors(new_connectors)
		self.cm.add_resources([
			Resource('spi', 0,
				Subsignal('ss',   Pins('1', conn = ('pmod_extension', 0))),
				Subsignal('clk',  Pins('2', conn = ('pmod_extension', 0))),
				Subsignal('miso', Pins('3', conn = ('pmod_extension', 0))),
				Subsignal('mosi', Pins('4', conn = ('pmod_extension', 0))),
			)
		])
		self.cm.request('spi', 0)
		self.assertEqual(list(self.cm.iter_port_constraints()), [
			('spi_0__ss__io',   ['B0'], {}),
			('spi_0__clk__io',  ['B1'], {}),
			('spi_0__miso__io', ['B2'], {}),
			('spi_0__mosi__io', ['B3'], {}),
		])

	def test_request_clock(self):
		clk100 = self.cm.request('clk100', 0)
		clk50 = self.cm.request('clk50', 0, dir = 'i')
		clk100_port_p, clk100_port_n, clk50_port = self.cm.iter_ports()
		self.assertEqual(list(self.cm.iter_clock_constraints()), [
			(clk100.i, clk100_port_p, 100e6),
			(clk50.i, clk50_port, 50e6)
		])

	def test_add_clock(self):
		i2c = self.cm.request('i2c')
		self.cm.add_clock_constraint(i2c.scl.o, 100e3)
		self.assertEqual(list(self.cm.iter_clock_constraints()), [
			(i2c.scl.o, None, 100e3)
		])

	def test_wrong_resources(self):
		with self.assertRaisesRegex(TypeError, r'^Object \'wrong\' is not a Resource$'):
			self.cm.add_resources(['wrong'])

	def test_wrong_resources_duplicate(self):
		with self.assertRaisesRegex(
			NameError, (
				r'^Trying to add \(resource user_led 0 \(pins o A1\)\), but '
				r'\(resource user_led 0 \(pins o A0\)\) has the same name and number$'
			)
		):
			self.cm.add_resources([Resource('user_led', 0, Pins('A1', dir = 'o'))])

	def test_wrong_connectors(self):
		with self.assertRaisesRegex(TypeError, r'^Object \'wrong\' is not a Connector$'):
			self.cm.add_connectors(['wrong'])

	def test_wrong_connectors_duplicate(self):
		with self.assertRaisesRegex(
			NameError, (
				r'^Trying to add \(connector pmod 0 1=>1 2=>2\), but '
				r'\(connector pmod 0 1=>B0 2=>B1 3=>B2 4=>B3\) has the same name and number$'
			)
		):
			self.cm.add_connectors([Connector('pmod', 0, '1 2')])

	def test_wrong_lookup(self):
		with self.assertRaisesRegex(
			ResourceError,
			r'^Resource user_led#1 does not exist$'
		):
			self.cm.lookup('user_led', 1)

	def test_wrong_clock_signal(self):
		with self.assertRaisesRegex(TypeError, r'^Object None is not a Signal$'):
			self.cm.add_clock_constraint(None, 10e6)

	def test_wrong_clock_frequency(self):
		with self.assertRaisesRegex(TypeError, r'^Frequency must be a number, not None$'):
			self.cm.add_clock_constraint(Signal(), None)

	def test_wrong_request_duplicate(self):
		with self.assertRaisesRegex(
			ResourceError,
			r'^Resource user_led#0 has already been requested$'
		):
			self.cm.request('user_led', 0)
			self.cm.request('user_led', 0)

	def test_wrong_request_duplicate_physical(self):
		self.cm.add_resources([
			Resource('clk20', 0, Pins('H1', dir = 'i')),
		])
		self.cm.request('clk100', 0)
		with self.assertRaisesRegex(
			ResourceError, (
				r'^Resource component clk20_0 uses physical pin H1, but it is already '
				r'used by resource component clk100_0 that was requested earlier$'
			)):
			self.cm.request('clk20', 0)

	def test_wrong_request_with_dir(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Direction must be one of \'i\', \'o\', \'oe\', \'io\', or \'-\', '
				r'not \'wrong\'$'
			)
		):
			self.cm.request('user_led', 0, dir = 'wrong')

	def test_wrong_request_with_dir_io(self):
		with self.assertRaisesRegex(
			ValueError, (
				r'^Direction of \(pins o A0\) cannot be changed from \'o\' to \'i\'; direction '
				r'can be changed from \'io\' to \'i\', \'o\', or \'oe\', or from anything '
				r'to \'-\'$'
			)
		):
			self.cm.request('user_led', 0, dir = 'i')

	def test_wrong_request_with_dir_dict(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Directions must be a dict, not \'i\', because \(resource i2c 0 \(subsignal scl '
				r'\(pins o N10\)\) \(subsignal sda \(pins io N11\)\)\) '
				r'has subsignals$'
			)
		):
			self.cm.request('i2c', 0, dir = 'i')

	def test_wrong_request_with_wrong_xdr(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Data rate of \(pins o A0\) must be a non-negative integer, not -1$'
		):
			self.cm.request('user_led', 0, xdr = -1)

	def test_wrong_request_with_xdr_dict(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Data rate must be a dict, not 2, because \(resource i2c 0 \(subsignal scl '
			r'\(pins o N10\)\) \(subsignal sda \(pins io N11\)\)\) '
			r'has subsignals$'
		):
			self.cm.request('i2c', 0, xdr = 2)

	def test_wrong_clock_constraint_twice(self):
		clk100 = self.cm.request('clk100')
		with self.assertRaisesRegex(
			ValueError, (
				r'^Cannot add clock constraint on \(sig clk100_0__i\), which is already '
				r'constrained to 100000000\.0 Hz$'
			)
		):
			self.cm.add_clock_constraint(clk100.i, 1e6)



# SPDX-License-Identifier: BSD-2-Clause

from typing         import Optional

from ....           import Cat, Elaboratable, Module, Signal
from ....util.units import log2_int
from ..memory       import MemoryMap
from ..wishbone     import Interface as WishboneInterface
from .              import Interface as CSRInterface

__all__ = (
	'WishboneCSRBridge',
)

class WishboneCSRBridge(Elaboratable):
	'''
	Wishbone to CSR bridge.

	A bus bridge for accessing CSR registers from Wishbone. This bridge supports any Wishbone
	data width greater or equal to CSR data width and performs appropriate address translation.

	Latency
	-------

	Reads and writes always take ``self.data_width // csr_bus.data_width + 1`` cycles to complete,
	regardless of the select inputs. Write side effects occur simultaneously with acknowledgement.

	Parameters
	----------
	csr_bus : :class:`..csr.Interface`
		CSR bus driven by the bridge.
	data_width : int or None
		Wishbone bus data width. If not specified, defaults to ``csr_bus.data_width``.
	name : str
		Window name. Optional.

	Attributes
	----------
	wb_bus : :class:`..wishbone.Interface`
		Wishbone bus provided by the bridge.

	'''

	def __init__(
		self, csr_bus: CSRInterface, *, data_width: Optional[int] = None, name: Optional[str] = None
	) -> None:
		if not isinstance(csr_bus, CSRInterface):
			raise ValueError(f'CSR bus must be an instance of CSRInterface, not {csr_bus!r}')
		if csr_bus.data_width not in (8, 16, 32, 64):
			raise ValueError(f'CSR bus data width must be one of 8, 16, 32, 64, not {csr_bus.data_width!r}')
		if data_width is None:
			data_width = csr_bus.data_width

		self.csr_bus = csr_bus
		self.wb_bus  = WishboneInterface(
			addr_width = max(0, csr_bus.addr_width - log2_int(data_width // csr_bus.data_width)),
			data_width = data_width,
			granularity = csr_bus.data_width,
			name = 'wb'
		)

		wb_map = MemoryMap(
			addr_width = csr_bus.addr_width,
			data_width = csr_bus.data_width,
			name = name
		)
		# Since granularity of the Wishbone interface matches the data width of the CSR bus,
		# no width conversion is performed, even if the Wishbone data width is greater.
		wb_map.add_window(self.csr_bus.memory_map)
		self.wb_bus.memory_map = wb_map

	def elaborate(self, platform) -> Module:
		csr_bus = self.csr_bus
		wb_bus  = self.wb_bus

		m = Module()

		cycle = Signal(range(len(wb_bus.sel) + 1))
		m.d.comb += csr_bus.addr.eq(Cat(cycle[:log2_int(len(wb_bus.sel))], wb_bus.adr))

		with m.If(wb_bus.cyc & wb_bus.stb):
			with m.Switch(cycle):
				def segment(index):
					return slice(index * wb_bus.granularity, (index + 1) * wb_bus.granularity)

				for index, sel_index in enumerate(wb_bus.sel):
					with m.Case(index):
						if index > 0:
							# CSR reads are registered, and we need to re-register them.
							m.d.sync += wb_bus.dat_r[segment(index - 1)].eq(csr_bus.r_data)
						m.d.comb += csr_bus.r_stb.eq(sel_index & ~wb_bus.we)
						m.d.comb += csr_bus.w_data.eq(wb_bus.dat_w[segment(index)])
						m.d.comb += csr_bus.w_stb.eq(sel_index & wb_bus.we)
						m.d.sync += cycle.eq(index + 1)

				with m.Default():
					m.d.sync += wb_bus.dat_r[segment(index)].eq(csr_bus.r_data)
					m.d.sync += wb_bus.ack.eq(1)

		with m.If(wb_bus.ack):
			m.d.sync += cycle.eq(0)
			m.d.sync += wb_bus.ack.eq(0)

		return m



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Module, Signal, Cat
from torii.back import verilog


class ALU(Elaboratable):
	def __init__(self, width: int):
		self.sel = Signal(2)
		self.a   = Signal(width)
		self.b   = Signal(width)
		self.o   = Signal(width)
		self.co  = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()
		with m.If(self.sel == 0b00):
			m.d.comb += self.o.eq(self.a | self.b)
		with m.Elif(self.sel == 0b01):
			m.d.comb += self.o.eq(self.a & self.b)
		with m.Elif(self.sel == 0b10):
			m.d.comb += self.o.eq(self.a ^ self.b)
		with m.Else():
			m.d.comb += Cat(self.o, self.co).eq(self.a - self.b)
		return m


if __name__ == '__main__':
	alu = ALU(width = 16)

	print(verilog.convert(alu, ports = [alu.sel, alu.a, alu.b, alu.o, alu.co]))



# SPDX-License-Identifier: BSD-2-Clause

from abc       import abstractproperty
from fractions import Fraction
import re

from ...hdl     import *
from ...lib.cdc import ResetSynchronizer
from ...build   import *

# Acknowledgments:
#   Parts of this file originate from https://github.com/tcjie/Gowin


__all__ = (
	'GowinPlatform',
)


class GowinPlatform(TemplatedPlatform):
	'''
	.. rubric:: Apicula toolchain

	Required tools:
		* ``yosys``
		* ``nextpnr-gowin``
		* ``gowin_pack``

	The environment is populated by running the script specified in the environment variable
	``AMARANTH_ENV_APICULA``, if present.

	Build products:
		* ``{{name}}.fs``: binary bitstream.

	.. rubric:: Gowin toolchain

	Required tools:
		* ``gw_sh``

	The environment is populated by running the script specified in the environment variable
	``AMARANTH_ENV_GOWIN``, if present.

	Build products:
		* ``{{name}}.fs``: binary bitstream.
	'''

	toolchain = None # selected when creating platform

	part   = abstractproperty()
	family = abstractproperty()

	def parse_part(self):
		# These regular expressions match all >900 parts of Gowin device_info.csv
		reg_series    = r'(GW[12]{1}[AN]{1}[EFNRSZ]{0,3})-'
		reg_voltage   = r'(ZV|EV|LV|LX|UV|UX)'
		reg_size      = r'(1|2|4|9|18|55)'
		reg_subseries = r'(?:(B|C|S|X|P5)?)'
		reg_package   = r'((?:PG|UG|EQ|LQ|MG|M|QN|CS|FN)(?:\d+)(?:P?)(?:A|E|M|CF|C|D|G|H|F|S|T|U|X)?)'
		reg_speed     = r'((?:C\d{1}/I\d{1})|ES|A\d{1}|I\d{1})'

		match = re.match(
			f'{reg_series}{reg_voltage}{reg_size}{reg_subseries}{reg_package}{reg_speed}$',
			self.part
		)

		if not match:
			raise ValueError(f'Supplied part name \'{self.part}\' is invalid!')

		self.series    = match.group(1)
		self.voltage   = match.group(2)
		self.size      = match.group(3)
		self.subseries = match.group(4) or ''
		self.package   = match.group(5)
		self.speed     = match.group(6)

		match = re.match(f'{reg_series}{reg_size}{reg_subseries}$', self.family)
		if not match:
			raise ValueError(f'Supplied device family name \'{self.family}\' is invalid')

		self.series_f    = match.group(1)
		self.size_f      = match.group(2)
		self.subseries_f = match.group(3) or ''

		# subseries_f is usually more reliable than subseries.

		if self.series != self.series_f:
			raise ValueError(
				'Series extracted from supplied part name does not match supplied family series'
			)
		if self.size != self.size_f:
			raise ValueError(
				'Size extracted from supplied part name does not match supplied family size'
			)

	# _chipdb_device is tied to available chipdb-*.bin files of nextpnr-gowin
	@property
	def _chipdb_device(self):
		# GW1NR series does not have its own chipdb file, but works with GW1N
		if self.series == 'GW1NR':
			return f'GW1N-{self.size}{self.subseries_f}'
		return self.family

	_dev_osc_mapping = {
		'GW1N-1'    : 'OSCH',
		'GW1N-1P5'  : 'OSCO',
		'GW1N-1P5B' : 'OSCO',
		'GW1N-1S'   : 'OSCH',
		'GW1N-2'    : 'OSCO',
		'GW1N-2B'   : 'OSCO',
		'GW1N-4'    : 'OSC',
		'GW1N-4B'   : 'OSC',
		'GW1N-9'    : 'OSC',
		'GW1N-9C'   : 'OSC',
		'GW1NR-1'   : 'OSCH',
		'GW1NR-2'   : 'OSCO',
		'GW1NR-2B'  : 'OSCO',
		'GW1NR-4'   : 'OSC',
		'GW1NR-4B'  : 'OSC',
		'GW1NR-9'   : 'OSC',
		'GW1NR-9C'  : 'OSC',
		'GW1NRF-4B' : 'OSC',
		'GW1NS-2'   : 'OSCF',
		'GW1NS-2C'  : 'OSCF',
		'GW1NS-4'   : 'OSCZ',
		'GW1NS-4C'  : 'OSCZ',
		'GW1NSE-2C' : 'OSCF',
		'GW1NSER-4C': 'OSCZ',
		'GW1NSR-2'  : 'OSCF',
		'GW1NSR-2C' : 'OSCF',
		'GW1NSR-4'  : 'OSCZ',
		'GW1NSR-4C' : 'OSCZ',
		'GW1NZ-1'   : 'OSCZ',
		'GW1NZ-1C'  : 'OSCZ',
		'GW2A-18'   : 'OSC',
		'GW2A-18C'  : 'OSC',
		'GW2A-55'   : 'OSC',
		'GW2A-55C'  : 'OSC',
		'GW2AN-18X' : 'OSCW',
		'GW2AN-55C' : 'OSC',
		'GW2AN-9X'  : 'OSCW',
		'GW2ANR-18C': 'OSC',
		'GW2AR-18'  : 'OSC',
		'GW2AR-18C' : 'OSC'
	}

	@property
	def _osc_type(self):
		if self.family in self._dev_osc_mapping:
			return self._dev_osc_mapping[self.family]
		raise NotImplementedError(
			f'Device family {self.family} does not have an assigned oscillator type'
		)

	@property
	def _osc_base_freq(self):
		osc = self._osc_type
		if osc == 'OSC':
			if self.speed == 4 and self.subseries_f in ('B', 'D'):
				return 210_000_000
			else:
				return 250_000_000
		elif osc in ('OSCZ', 'OSCO'):
			if self.series == 'GW1NSR' and self.speed == 'C7/I6':
				return 260_000_000
			else:
				return 250_000_000
		elif osc in ('OSCF', 'OSCH'):
			return 240_000_000
		elif osc == 'OSCW':
			return 200_000_000
		else:
			assert False

	@property
	def _osc_div(self):
		div_min  = 2
		div_max  = 128
		div_step = 2
		div_frac = Fraction(self._osc_base_freq, self.osc_frequency)

		if div_frac.denominator != 1 or div_frac not in range(div_min, div_max, div_step):
			raise ValueError(
				'On-chip oscillator frequency (platform.osc_frequency) must be chosen such that '
				f'the oscillator divider, calculated as ({div_frac.numerator}/{div_frac.denominator}), '
				f'is an integer between {div_min} and {div_max} in '
				f'steps of {div_step}'
			)

		return div_frac.numerator

	# Common templates

	_common_file_templates = {
		'{{name}}.cst': r'''
			// {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				IO_LOC "{{port_name}}" {{pin_name}};
				{% for attr_name, attr_value in attrs.items() -%}
					IO_PORT "{{port_name}}" {{attr_name}}={{attr_value}};
				{% endfor %}
			{% endfor %}
		''',
	}

	# Apicula templates

	_apicula_required_tools = [
		'yosys',
		'nextpnr-gowin',
		'gowin_pack'
	]
	_apicula_file_templates = {
		**TemplatedPlatform.build_script_templates,
		**_common_file_templates,
		'{{name}}.il': r'''
			# {{autogenerated}}
			{{emit_rtlil()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.ys': r'''
			# {{autogenerated}}
			{% for file in platform.iter_files(".v") -%}
				read_verilog {{get_override("read_verilog_opts")|options}} {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".sv") -%}
				read_verilog -sv {{get_override("read_verilog_opts")|options}} {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".il") -%}
				read_ilang {{file}}
			{% endfor %}
			read_ilang {{name}}.il
			delete w:$verilog_initial_trigger
			{{get_override("script_after_read")|default("# (script_after_read placeholder)")}}
			synth_gowin {{get_override("synth_opts")|options}} -top {{name}} -json {{name}}.syn.json
			{{get_override("script_after_synth")|default("# (script_after_synth placeholder)")}}
		''',
	}
	_apicula_command_templates = [
		r'''
		{{invoke_tool("yosys")}}
			{{quiet("-q")}}
			{{get_override("yosys_opts")|options}}
			-l {{name}}.rpt
			{{name}}.ys
		''',
		r'''
		{{invoke_tool("nextpnr-gowin")}}
			{{quiet("--quiet")}}
			{{get_override("nextpnr_opts")|options}}
			--log {{name}}.tim
			--device {{platform.part}}
			--family {{platform._chipdb_device}}
			--json {{name}}.syn.json
			--cst {{name}}.cst
			--write {{name}}.pnr.json
		''',
		r'''
		{{invoke_tool("gowin_pack")}}
			-d {{platform._chipdb_device}}
			-o {{name}}.fs
			{{get_override("gowin_pack_opts")|options}}
			{{name}}.pnr.json
		'''
	]

	# Vendor toolchain templates

	_gowin_required_tools = [ 'gw_sh' ]
	_gowin_file_templates = {
		**TemplatedPlatform.build_script_templates,
		**_common_file_templates,
		'{{name}}.v': r'''
			/* {{autogenerated}} */
			{{emit_verilog()}}
		''',
		'{{name}}.tcl': r'''
			# {{autogenerated}}
			{% for file in platform.iter_files(".v",".sv",".vhd",".vhdl") -%}
				add_file {{file}}
			{% endfor %}
			add_file -type verilog {{name}}.v
			add_file -type cst {{name}}.cst
			add_file -type sdc {{name}}.sdc
			set_device -name {{platform.family}} {{platform.part}}
			set_option -verilog_std v2001 -print_all_synthesis_warning 1 -show_all_warn 1
			{{get_override("add_options")|default("# (add_options placeholder)")}}
			run all
			file delete -force {{name}}.fs
			file copy -force impl/pnr/project.fs {{name}}.fs
		''',
		'{{name}}.sdc': r'''
		// {{autogenerated}}
		{% for net_signal,port_signal,frequency in platform.iter_clock_constraints() -%}
			create_clock -name {{port_signal.name|tcl_escape}} -period {{1000000000/frequency}}  [get_ports {{port_signal.name|tcl_escape}}]
		{% endfor %}
		{{get_override("add_constraints")|default("# (add_constraints placeholder)")}}
		''',
	}
	_gowin_command_templates = [
		r'''
		{{invoke_tool("gw_sh")}}
			{{name}}.tcl
		'''
	]

	def __init__(self, *, toolchain = 'Apicula'):
		super().__init__()

		assert toolchain in ('Apicula', 'Gowin')
		self.toolchain = toolchain

		self.parse_part()

	@property
	def required_tools(self):
		if self.toolchain == 'Apicula':
			return self._apicula_required_tools
		elif self.toolchain == 'Gowin':
			return self._gowin_required_tools
		assert False

	@property
	def file_templates(self):
		if self.toolchain == 'Apicula':
			return self._apicula_file_templates
		elif self.toolchain == 'Gowin':
			return self._gowin_file_templates
		assert False

	@property
	def command_templates(self):
		if self.toolchain == 'Apicula':
			return self._apicula_command_templates
		elif self.toolchain == 'Gowin':
			return self._gowin_command_templates
		assert False

	def add_clock_constraint(self, clock, frequency):
		super().add_clock_constraint(clock, frequency)
		clock.attrs['keep'] = 'true'

	@property
	def default_clk_constraint(self):
		if self.default_clk == 'OSC':
			if not hasattr(self, 'osc_frequency'):
				raise AttributeError(
					'Using the on-chip oscillator as the default clock source requires '
					'the platform.osc_frequency attribute to be set'
				)
			return Clock(self.osc_frequency)

		# Use the defined Clock resource.
		return super().default_clk_constraint

	def create_missing_domain(self, name):
		if name == 'sync' and self.default_clk is not None:
			m = Module()

			if self.default_clk == 'OSC':
				clk_i = Signal()
				if self._osc_type == 'OSCZ':
					m.submodules += Instance(
						self._osc_type,
						p_FREQ_DIV = self._osc_div,
						i_OSCEN    = Const(1),
						o_OSCOUT   = clk_i
					)
				elif self._osc_type == 'OSCO':
					# TODO: Make use of regulator configurable
					m.submodules += Instance(
						self._osc_type,
						p_REGULATOR_EN = Const(1),
						p_FREQ_DIV     = self._osc_div,
						i_OSCEN        = Const(1),
						o_OSCOUT       = clk_i
					)
				elif self._osc_type == 'OSCF':
					m.submodules += Instance(
						self._osc_type,
						p_FREQ_DIV  = self._osc_div,
						o_OSCOUT30M = None,
						o_OSCOUT    = clk_i
					)
				else:
					m.submodules += Instance(
						self._osc_type,
						p_FREQ_DIV = self._osc_div,
						o_OSCOUT   = clk_i
					)

			else:
				clk_i = self.request(self.default_clk).i

			if self.default_rst is not None:
				rst_i = self.request(self.default_rst).i
			else:
				rst_i = Const(0)

			m.submodules.reset_sync = ResetSynchronizer(rst_i, domain = 'sync')
			m.domains += ClockDomain('sync')
			m.d.comb += ClockSignal('sync').eq(clk_i)

			return m

	def _get_xdr_buffer(self, m, pin, i_invert = False, o_invert = False):

		def get_ireg(clk, d, q):
			for bit in range(len(q)):
				m.submodules += Instance(
					'DFF',
					i_CLK = clk,
					i_D   = d[bit],
					o_Q   = q[bit],
				)

		def get_oreg(clk, d, q):
			for bit in range(len(q)):
				m.submodules += Instance(
					'DFF',
					i_CLK = clk,
					i_D   = d[bit],
					o_Q   = q[bit]
				)

		def get_iddr(clk, d, q0, q1):
			for bit in range(len(d)):
				m.submodules += Instance(
					'IDDR',
					i_CLK = clk,
					i_D   = d[bit],
					o_Q0  = q0[bit],
					o_Q1  = q1[bit]
				)

		def get_oddr(clk, d0, d1, q):
			for bit in range(len(q)):
				m.submodules += Instance(
					'ODDR',
					p_TXCLK_POL = 0, # default -> Q1 changes on posedge of CLK
					i_CLK = clk,
					i_D0 = d0[bit],
					i_D1 = d1[bit],
					o_Q0 = q[bit]
				)


		def get_oeddr(clk, d0, d1, tx, q0, q1):
			for bit in range(len(q0)):
				m.submodules += Instance(
					'ODDR',
					p_TXCLK_POL = 0, # default -> Q1 changes on posedge of CLK
					i_CLK       = clk,
					i_D0        = d0[bit],
					i_D1        = d1[bit],
					i_TX        = tx,
					o_Q0        = q0[bit],
					o_Q1        = q1
				)

		def get_ineg(y, invert):
			if invert:
				a = Signal.like(y, name_suffix = '_n')
				m.d.comb += y.eq(~a)
				return a
			else:
				return y

		def get_oneg(a, invert):
			if invert:
				y = Signal.like(a, name_suffix = '_n')
				m.d.comb += y.eq(~a)
				return y
			else:
				return a


		if 'i' in pin.dir:
			if pin.xdr < 2:
				pin_i = get_ineg(pin.i, i_invert)
			elif pin.xdr == 2:
				pin_i0 = get_ineg(pin.i0, i_invert)
				pin_i1 = get_ineg(pin.i1, i_invert)
		if 'o' in pin.dir:
			if pin.xdr < 2:
				pin_o = get_oneg(pin.o, o_invert)
			elif pin.xdr == 2:
				pin_o0 = get_oneg(pin.o0, o_invert)
				pin_o1 = get_oneg(pin.o1, o_invert)

		i = o = t = None

		if 'i' in pin.dir:
			i = Signal(pin.width, name = f'{pin.name}_xdr_i')
		if 'o' in pin.dir:
			o = Signal(pin.width, name = f'{pin.name}_xdr_o')
		if pin.dir in ('oe', 'io'):
			t = Signal(1,         name = f'{pin.name}_xdr_t')

		if pin.xdr == 0:
			if 'i' in pin.dir:
				i = pin_i
			if 'o' in pin.dir:
				o = pin_o
			if pin.dir in ('oe', 'io'):
				t = ~pin.oe
		elif pin.xdr == 1:
			if 'i' in pin.dir:
				get_ireg(pin.i_clk, i, pin_i)
			if 'o' in pin.dir:
				get_oreg(pin.o_clk, pin_o, o)
			if pin.dir in ('oe', 'io'):
				get_oreg(pin.o_clk, ~pin.oe, t)
		elif pin.xdr == 2:
			if 'i' in pin.dir:
				get_iddr(pin.i_clk, i, pin_i0, pin_i1)
			if pin.dir in ('o', ):
				get_oddr(pin.o_clk, pin_o0, pin_o1, o, )
			if pin.dir in ('oe', 'io'):
				get_oeddr(pin.o_clk, pin_o0, pin_o1, ~pin.oe, o, t)
		else:
			assert False

		return (i, o, t)

	def get_input(self, pin, port, attrs, invert):
		self._check_feature(
			'single-ended input', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)
		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'IBUF',
				i_I = port.io[bit],
				o_O = i[bit]
			)
		return m

	def get_output(self, pin, port, attrs, invert):
		self._check_feature(
			'single-ended output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)
		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, port.io, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OBUF',
				i_I = o[bit],
				o_O = port.io[bit]
			)
		return m

	def get_tristate(self, pin, port, attrs, invert):
		self._check_feature(
			'single-ended tristate', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)
		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'TBUF',
				i_OEN = t,
				i_I   = o[bit],
				o_O   = port.io[bit]
			)
		return m

	def get_input_output(self, pin, port, attrs, invert):
		self._check_feature(
			'single-ended input/output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)
		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'IOBUF',
				i_OEN = t,
				i_I   = o[bit],
				o_O   = i[bit],
				io_IO = port.io[bit]
			)
		return m

	def get_diff_input(self, pin, port, attrs, invert):
		self._check_feature(
			'differential input', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)
		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert)
		for bit in range(pin.wodth):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'TLVDS_IBUF',
				i_I  = port.p[bit],
				i_IB = port.n[bit],
				o_O  = i[bit]
			)
		return m

	def get_diff_output(self, pin, port, attrs, invert):
		self._check_feature(
			'differential output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)
		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'TLVDS_OBUF',
				i_I  = o[bit],
				o_O  = port.p[bit],
				o_OB = port.n[bit],
			)
		return m

	def get_diff_tristate(self, pin, port, attrs, invert):
		self._check_feature(
			'differential tristate', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)
		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'TLVDS_TBUF',
				i_OEN = t,
				i_I   = o[bit],
				o_O   = port.p[bit],
				o_OB  = port.n[bit]
			)
		return m

	def get_diff_input_output(self, pin, port, attrs, invert):
		self._check_feature(
			'differential input/output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)
		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'TLVDS_IOBUF',
				i_OEN  = t,
				i_I    = o[bit],
				o_O    = i[bit],
				io_IO  = port.p[bit],
				io_IOB = port.n[bit]
			)
		return m
